﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using System.IO;
using System.Windows.Forms;
using System.Net;
using Xceed.Zip;
using Xceed.FileSystem;
using HaloMap;

namespace Gravemind
{
    public class Update
    {
        public event DownloadProgressChangedEventHandler DownloadProgress;
        public event System.ComponentModel.AsyncCompletedEventHandler DownloadComplete;
        public event ByteProgressionEventHandler ByteProgression;

        public static void CheckUpdate(Main m)
        {
            try
            {
                //Read Xml
                WebClient wc = new WebClient();
                wc.DownloadFile(new Uri(Globals.ServerPath + Globals.UpdateFile), Application.StartupPath + "\\Update.Xml");
                XmlTextReader xr = new XmlTextReader(new FileStream(Application.StartupPath + "\\Update.xml", FileMode.Open, FileAccess.Read, FileShare.Read));
                xr.ReadStartElement("Update");
                double Version = Convert.ToDouble(xr.ReadElementString("Version"));
                xr.Close();

                //Compare Version
                if (Version > Globals.Version)
                {
                    if (!Globals.UpdateWithoutPrompt)
                    {
                        if (MessageBox.Show("There is an update available, Would you like to update?", "Update Available", MessageBoxButtons.YesNo) == DialogResult.Yes)
                        {
                            Updater u = new Updater();
                            u.MdiParent = m;
                            u.Show();
                        }
                    }
                    else
                    {
                        Updater u = new Updater();
                        u.MdiParent = m;
                        u.Show();
                    }
                }
            }
            catch (Exception)
            {
                MessageBox.Show("Error checking for Update!");
                return;
            }
        }

        public void DownloadUpdate()
        {
            //Intialize
            WebClient wc = new WebClient();
            wc.DownloadProgressChanged += new DownloadProgressChangedEventHandler(DownloadProgress);
            wc.DownloadFileCompleted += new System.ComponentModel.AsyncCompletedEventHandler(DownloadComplete);

            //Download
            wc.DownloadFileAsync( new Uri(Globals.ServerPath + Globals.UpdateName), Application.StartupPath + "\\Update.zip");
        }

        public void InstallUpdate()
        {
            AbstractFile zipFile = new DiskFile(Application.StartupPath + "\\Update.zip");
            ZipArchive archive = new ZipArchive(zipFile);

            foreach (FileSystemItem childItem in archive.GetItems(true))
            {
                string name = childItem.FullName;
                if (File.Exists(Application.StartupPath + name))
                {
                     File.Move(Application.StartupPath + name, Application.StartupPath + name + ".old");
                }
            }

            AbstractFolder dest = new DiskFolder(Application.StartupPath);

            ZipEvents events = new ZipEvents();
            events.ByteProgression += ByteProgression;

            archive.CopyFilesTo(events, null, dest, true, true, "*.*");
        }

        public static void Clean()
        {
            //Clean
            DirectoryInfo di = new DirectoryInfo(Application.StartupPath);
            Clean(di);
        }

        public static void Clean(DirectoryInfo di)
        {
            DirectoryInfo[] ds = di.GetDirectories();
            for (int i = 0; i < ds.Length; i++)
            {
                Clean(ds[i]);
            }

            FileInfo[] fs = di.GetFiles();
            for (int i = 0; i < fs.Length; i++)
            {
                string ext = fs[i].Extension;
                if (ext == ".old")
                {
                    fs[i].Delete();
                }
            }
        }
    }
}
