﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HaloMap.BSP;
using System.IO;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace HaloMap.BSP
{
    public class DetailObject
    {
        public int VertexCount, FaceCount;
        public int RawPointerOffset, RawOffset, RawSize, RawHeaderSize, RawSizeWithoutHeader;
        public List<Resource> Resources = new List<Resource>();
        public BSP1 Raw = new BSP1();
    }

    public class BSPPermutation
    {
        public short VertexCount, TriangleCount, PartCount;
        public int RawPointerOffset, RawOffset, RawSize, RawHeaderSize, RawWithoutHeader;
        public List<Resource> Resources = new List<Resource>();
        public BSP2 Raw = new BSP2();
    }
    public class Resource
    {
        public short PrimaryLocator, ChunkSize;
        public int Size, Offset;
    }

    public class WaterDefinition
    {
        public TagReference Shader = new TagReference();
        public int RawPointerOffset, RawOffset, RawSize, RawHeaderSize, RawWithoutHeader;
        public List<Resource> Resources = new List<Resource>();
    }

    public class Decorator
    {
        public List<CacheBlock> CacheBlocks = new List<CacheBlock>();
    }
    public class CacheBlock
    {
        public int RawPointerOffset, RawOffset, RawSize, RawHeaderSize, RawWithoutHeader;
        public List<Resource> Resources = new List<Resource>();
    }

    #region Generic Classes
    public class MeshInfo
    {
        public List<short> ShaderNumber = new List<short>(),
            IndiceStart = new List<short>(),
            IndiceCount = new List<short>();
    }
    public class UnknownBlock
    {
        public List<int> Unknown1 = new List<int>(), Unknown2 = new List<int>();
    }
    public class Indices
    {
        public List<short> Indice = new List<short>();
    }
    public class VerticeBlock
    {
        public List<Vector3> Vertice = new List<Vector3>();
    }
    public class UVData
    {
        public List<Vector2> UVs = new List<Vector2>();
    }
    public class NormalsBlock
    {
        public List<Vector3> Normals = new List<Vector3>(),
            Binormals = new List<Vector3>(),
            Tangents = new List<Vector3>();
    }
    public class LightmapUVData
    {
        public List<Vector2> UVs = new List<Vector2>();
    }
    #endregion

    public class BSP1
    {
        #region Fields
        public string BlockHeader;
        public int DataSize, MeshInfoChunkCount, IndiceCount, VerticeCount;
        public MeshInfo Mesh = new MeshInfo();
        public UnknownBlock Unknown = new UnknownBlock();
        public Indices Indices = new Indices();
        public VerticeBlock Vertices = new VerticeBlock();
        public UVData UVData = new UVData();
        public NormalsBlock Normals = new NormalsBlock();
        public LightmapUVData LightmapUV = new LightmapUVData();
        #endregion

        #region Methods
        public void Read(Map map, int Offset, List<Resource> Res)
        {
            map.br.BaseStream.Position = Offset;
            BlockHeader = new string(map.br.ReadChars(4));
            DataSize = map.br.ReadInt32();
            MeshInfoChunkCount = map.br.ReadInt32();
            map.br.BaseStream.Position = Offset + 40;
            IndiceCount = map.br.ReadInt32();

            //Mesh Info
            map.br.BaseStream.Position = Offset + 76 + Res[0].Offset;
            for (int i = 0; i < IndiceCount; i++)
            {
                Mesh.ShaderNumber.Add(map.br.ReadInt16());
                Mesh.IndiceStart.Add(map.br.ReadInt16());
                Mesh.IndiceCount.Add(map.br.ReadInt16());
            }

            //Unknown
            map.br.BaseStream.Position = Offset + 76 + Res[1].Offset;
            for (int i = 0; i < Res[1].Size / 8; i++)
            {
                Unknown.Unknown1.Add(map.br.ReadInt32());
                Unknown.Unknown2.Add(map.br.ReadInt32());
            }

            //Indices
            map.br.BaseStream.Position = Offset + 76 + Res[2].Offset;
            for (int i = 0; i < IndiceCount; i++)
            {
                Indices.Indice.Add(map.br.ReadInt16());
            }

            //Vertices
            map.br.BaseStream.Position = Offset + 76 + Res[5].Offset;
            for (int i = 0; i < Res[5].Size / 12; i++)
            {
                Vertices.Vertice.Add(new Vector3(map.br.ReadSingle(), map.br.ReadSingle(), map.br.ReadSingle()));
            }

            //UV Data
            map.br.BaseStream.Position = Offset + 76 + Res[6].Offset;
            for (int i = 0; i < Res[5].Size / 12; i++)
            {
                UVData.UVs.Add(new Vector2(map.br.ReadSingle(), map.br.ReadSingle()));
            }

            //Normals, Binormals, Tangents
            map.br.BaseStream.Position = Offset + 76 + Res[7].Offset;
            for (int i = 0; i < Res[5].Size / 12; i++)
            {
                Normals.Normals.Add(Bsp.DecompressNormal(map.br.ReadInt32()));
                Normals.Binormals.Add(Bsp.DecompressNormal(map.br.ReadInt32()));
                Normals.Tangents.Add(Bsp.DecompressNormal(map.br.ReadInt32()));
            }

            //Lightmap UV Data
            map.br.BaseStream.Position = Offset + 76 + Res[8].Offset;
            for (int i = 0; i < Res[5].Size / 12; i++)
            {
                LightmapUV.UVs.Add(new Vector2(Bsp.DecompressVertice(Convert.ToSingle(map.br.ReadInt16()), -1, 1), Bsp.DecompressVertice(Convert.ToSingle(map.br.ReadInt16()), -1, 1)));
            }
        }
        #endregion
    }

    public class BSP2
    {
        #region Fields
        public string BlockHeader;
        public int DataSize, MeshInfoChunkCount, IndiceCount, VerticeCount;
        public MeshInfo Mesh = new MeshInfo();
        public UnknownBlock Unknown = new UnknownBlock();
        public Indices Indices = new Indices();
        public VerticeBlock Vertices = new VerticeBlock();
        public UVData UVData = new UVData();
        public NormalsBlock Normals = new NormalsBlock();
        public LightmapUVData LightmapUV = new LightmapUVData();
        #endregion

        #region Methods
        public void Read(Map map, int Offset, List<Resource> Res)
        {
            map.br.BaseStream.Position = Offset;
            BlockHeader = new string(map.br.ReadChars(4));
            DataSize = map.br.ReadInt32();
            MeshInfoChunkCount = map.br.ReadInt32();
            map.br.BaseStream.Position = Offset + 40;
            IndiceCount = map.br.ReadInt32();

            //Mesh Info
            map.br.BaseStream.Position = Offset + 76 + Res[0].Offset;
            for (int i = 0; i < Res[0].Size / 6; i++)
            {
                Mesh.ShaderNumber.Add(map.br.ReadInt16());
                Mesh.IndiceStart.Add(map.br.ReadInt16());
                Mesh.IndiceCount.Add(map.br.ReadInt16());
            }

            //Unknown
            map.br.BaseStream.Position = Offset + 76 + Res[1].Offset;
            for (int i = 0; i < Res[1].Size / 8; i++)
            {
                Unknown.Unknown1.Add(map.br.ReadInt32());
                Unknown.Unknown2.Add(map.br.ReadInt32());
            }

            //Indices
            map.br.BaseStream.Position = Offset + 76 + Res[2].Offset;
            for (int i = 0; i < Res[2].Size / 2; i++)
            {
                Indices.Indice.Add(map.br.ReadInt16());
            }

            //Vertices
            map.br.BaseStream.Position = Offset + 76 + Res[5].Offset;
            for (int i = 0; i < Res[5].Size / 12; i++)
            {
                Vertices.Vertice.Add(new Vector3(map.br.ReadSingle(), map.br.ReadSingle(), map.br.ReadSingle()));
            }

            //UV Data
            map.br.BaseStream.Position = Offset + 76 + Res[6].Offset;
            for (int i = 0; i < Res[6].Size / 8; i++)
            {
                UVData.UVs.Add(new Vector2(map.br.ReadSingle(), map.br.ReadSingle()));
            }

            //Normals, Binormals, Tangents
            map.br.BaseStream.Position = Offset + 76 + Res[7].Offset;
            for (int i = 0; i < Res[7].Size / 12; i++)
            {
                Normals.Normals.Add(Bsp.DecompressNormal(map.br.ReadInt32()));
                Normals.Binormals.Add(Bsp.DecompressNormal(map.br.ReadInt32()));
                Normals.Tangents.Add(Bsp.DecompressNormal(map.br.ReadInt32()));
            }

            //Lightmap UV Data
            map.br.BaseStream.Position = Offset + 76 + Res[8].Offset;
            for (int i = 0; i < Res[8].Size / 4; i++)
            {
                LightmapUV.UVs.Add(new Vector2(Bsp.DecompressVertice(Convert.ToSingle(map.br.ReadInt16()), -1, 1), Bsp.DecompressVertice(Convert.ToSingle(map.br.ReadInt16()), -1, 1)));
            }
        }
        #endregion
    }
}

namespace HaloMap
{
    public class Bsp
    {
        public int Magic;
        public Tag tag;
        public Lightmap Lightmap = new Lightmap();

        public List<DetailObject> DetailObjects = new List<DetailObject>();
        public List<BSPPermutation> Permutations = new List<BSPPermutation>();
        public List<WaterDefinition> WaterDefinitions = new List<WaterDefinition>();
        public List<BSP.Decorator> Decorators = new List<BSP.Decorator>();

        public void Read(Map map, Tag Tag)
        {
            this.tag = Tag;

            //Calculate Bsp Magic
            //map.br.BaseStream.Position = Tag.MetaOffset + 16;
            //Magic = map.br.ReadInt32() - 572;
            //BSP's don't use magics?

            //Detail Object Raw
            map.br.BaseStream.Position = Tag.MetaOffset + 156;
            Reflexive r = new Reflexive();
            r.Read(map, true);
            for (int i = 0; i < r.ChunkCount; i++)
            {
                DetailObject Object = new DetailObject();
                map.br.BaseStream.Position = r.Translation + (176 * i);
                Object.VertexCount = map.br.ReadInt32();
                Object.FaceCount = map.br.ReadInt32();
                map.br.BaseStream.Position += 32;
                Object.RawPointerOffset = (int)map.br.BaseStream.Position;
                Object.RawOffset = map.br.ReadInt32();
                Object.RawSize = map.br.ReadInt32();
                Object.RawHeaderSize = map.br.ReadInt32();
                Object.RawSizeWithoutHeader = map.br.ReadInt32();
                Reflexive Resource = new Reflexive();
                Resource.Read(map, true);
                for (int z = 0; z < Resource.ChunkCount; z++)
                {
                    HaloMap.BSP.Resource Res = new HaloMap.BSP.Resource();
                    map.br.BaseStream.Position = Resource.Translation + (12 * z) + 4;
                    Res.PrimaryLocator = map.br.ReadInt16();
                    Res.ChunkSize = map.br.ReadInt16();
                    Res.Size = map.br.ReadInt32();
                    Res.Offset = map.br.ReadInt32();
                    Object.Resources.Add(Res);
                }
                if (Object.RawOffset != -1)
                {
                    Object.Raw.Read(map, Object.RawOffset, Object.Resources);
                }
                DetailObjects.Add(Object);
            }

            //BSP Permutations
            map.br.BaseStream.Position = Tag.MetaOffset + 312;
            r.Read(map, true);
            for (int i = 0; i < r.ChunkCount; i++)
            {
                BSPPermutation Perm = new BSPPermutation();
                map.br.BaseStream.Position = r.Translation + (200 * i);
                Perm.VertexCount = map.br.ReadInt16();
                Perm.TriangleCount = map.br.ReadInt16();
                Perm.PartCount = map.br.ReadInt16();
                map.br.BaseStream.Position += 34;
                Perm.RawPointerOffset = (int)map.br.BaseStream.Position;
                Perm.RawOffset = map.br.ReadInt32();
                Perm.RawSize = map.br.ReadInt32();
                Perm.RawHeaderSize = map.br.ReadInt32();
                Perm.RawWithoutHeader = map.br.ReadInt32();
                Reflexive Resources = new Reflexive();
                Resources.Read(map, true);
                for (int x = 0; x < Resources.ChunkCount; x++)
                {
                    HaloMap.BSP.Resource Res = new HaloMap.BSP.Resource();
                    map.br.BaseStream.Position = Resources.Translation + (16 * x) + 4;
                    Res.PrimaryLocator = map.br.ReadInt16();
                    Res.ChunkSize = map.br.ReadInt16();
                    Res.Size = map.br.ReadInt32();
                    Res.Offset = map.br.ReadInt32();
                    Perm.Resources.Add(Res);
                }
                Perm.Raw.Read(map, Perm.RawOffset, Perm.Resources);
                Permutations.Add(Perm);
            }

            //Water Definitions
            map.br.BaseStream.Position = Tag.MetaOffset + 532;
            r.Read(map, true);
            for (int i = 0; i < r.ChunkCount; i++)
            {
                WaterDefinition Water = new WaterDefinition();
                map.br.BaseStream.Position = r.Translation + (172 * i);
                Water.Shader.Read(map, true);
                map.br.BaseStream.Position += 8;
                Water.RawPointerOffset = (int)map.br.BaseStream.Position;
                Water.RawOffset = map.br.ReadInt32();
                Water.RawSize = map.br.ReadInt32();
                Water.RawHeaderSize = map.br.ReadInt32();
                Water.RawWithoutHeader = map.br.ReadInt32();
                Reflexive Resources = new Reflexive();
                Resources.Read(map, true);
                for (int x = 0; x < Resources.ChunkCount; x++)
                {
                    HaloMap.BSP.Resource Res = new HaloMap.BSP.Resource();
                    map.br.BaseStream.Position = Resources.Translation + (16 * x) + 4;
                    Res.PrimaryLocator = map.br.ReadInt16();
                    Res.ChunkSize = map.br.ReadInt16();
                    Res.Size = map.br.ReadInt32();
                    Res.Offset = map.br.ReadInt32();
                    Water.Resources.Add(Res);
                }
                WaterDefinitions.Add(Water);
            }

            //Decorator Raw
            map.br.BaseStream.Position = Tag.MetaOffset + 564;
            r.Read(map, true);
            for (int i = 0; i < r.ChunkCount; i++)
            {
                BSP.Decorator Dec = new BSP.Decorator();
                map.br.BaseStream.Position = r.Translation + (48 * i) + 16;
                Reflexive Cache = new Reflexive();
                Cache.Read(map, true);
                for (int x = 0; x < Cache.ChunkCount; x++)
                {
                    CacheBlock Block = new CacheBlock();
                    map.br.BaseStream.Position = Cache.Translation + (44 * x);
                    Block.RawPointerOffset = (int)map.br.BaseStream.Position;
                    Block.RawOffset = map.br.ReadInt32();
                    Block.RawSize = map.br.ReadInt32();
                    Block.RawHeaderSize = map.br.ReadInt32();
                    Block.RawWithoutHeader = map.br.ReadInt32();
                    //Reflexive Resource = new Reflexive();
                    //Resource.Read(map, true);
                    //for (int z = 0; z < Resource.ChunkCount; z++)
                    //{
                    //    HaloMap.BSP.Resource Res = new HaloMap.BSP.Resource();
                    //    map.br.BaseStream.Position = Resource.Translation + (16 * z) + 4;
                    //    Res.PrimaryLocator = map.br.ReadInt16();
                    //    Res.ChunkSize = map.br.ReadInt16();
                    //    Res.Size = map.br.ReadInt32();
                    //    Res.Offset = map.br.ReadInt32();
                    //    Block.Resources.Add(Res);
                    //}
                    Dec.CacheBlocks.Add(Block);
                }
                Decorators.Add(Dec);
            }
        }

        public void ExportBSP(Map map, string Path)
        {
            //Convert BSP raw to Model raw
            for (int i = 0; i < DetailObjects.Count; i++)
            {
                BinaryWriter bw = new BinaryWriter(new FileStream(Path + "\\DetailObject[" + i.ToString() + "].raw", FileMode.Create, FileAccess.Write, FileShare.Write));
            }

            for (int i = 0; i < Permutations.Count; i++)
            {
                BinaryWriter bw = new BinaryWriter(new FileStream(Path + "\\Permutation[" + i.ToString() + "].raw", FileMode.Create, FileAccess.Write, FileShare.Write));
            }

            //Export Sky
        }

        public void LoadBSP(Map map, int Index)
        {

        }

        public static Vector3 DecompressNormal(int compressednormal)
        {
            int xx = compressednormal & 0x3ff;
            float xxx = Convert.ToSingle(xx) / Convert.ToSingle(0x3ff);
            int bitx = ((compressednormal >> 10) & 0x1);
            if (bitx == 1) { xxx = 1 - xxx; xxx = -xxx; }

            int yy = ((compressednormal >> 11) & 0x3ff);
            float yyy = Convert.ToSingle(yy) / Convert.ToSingle(0x3ff);
            int bity = ((compressednormal >> 21) & 0x1);
            if (bity == 1) { yyy = 1 - yyy; yyy = -yyy; }

            int zz = ((compressednormal >> 22) & 0x1ff);
            float zzz = Convert.ToSingle(zz) / Convert.ToSingle(0x1ff);
            int bitz = ((compressednormal >> 31) & 0x1);
            if (bitz == 1) { zzz = 1 - zzz; zzz = -zzz; }
            return new Vector3(xxx, yyy, zzz);
        }
        public static int CompressNormal(Vector3 normal)
        {
            bool negx = false;
            if (normal.X < 0) { negx = true; normal.X = -normal.X; }
            float bbb = normal.X * Convert.ToSingle(0x3ff);
            int b = Convert.ToInt32(bbb);
            if (negx) b = -b + 1 | 0x400;

            bool negy = false;
            if (normal.Y < 0) { negy = true; normal.Y = -normal.Y; }
            float bbb2 = normal.Y * Convert.ToSingle(0x3ff);
            int b2 = Convert.ToInt32(bbb2);
            if (negy) b2 = -b2 + 1 | 0x400;
            b2 <<= 11;

            bool negz = false;
            if (normal.Z < 0) { negz = true; normal.Z = -normal.Z; }
            float bbb3 = normal.Z * Convert.ToSingle(0x1ff);
            int b3 = Convert.ToInt32(bbb3);
            if (negz) b3 = -b3 + 1 | 0x200;
            b3 <<= 22;
            int result = b3 | b2 | b;
            return result;
        }

        public static float DecompressVertice(float input, float min, float max)
        {
            float percent = (input + 32768) / 65535;
            float result = (percent * (max - min)) + min;
            return result;
        }

        public static float CompressVertice(float input, float min, float max)
        {
            float result = input - min;
            result = (result / (max - min));
            result = result * 65535;
            result = result - 32768;
            return result;
        }
    }
}
