﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace HaloMap
{
    public partial class BitmapEditor : UserControl
    {
        public Map map;

        public BitmapEditor(Map map)
        {
            InitializeComponent();
            this.map = map;
        }

        public void Update(Map map)
        {
            map.UpdateProccess("Updating...");
            this.map = map;
            listBox1.Items.Clear();
            BitmapRaw.Read(map);

            for (int i = 0; i < BitmapRaw.RawChunks.Count; i++)
            {
                listBox1.Items.Add("Bitmap Chunk #" + i.ToString());
            }
            try { listBox1.SelectedIndex = 0; }
            catch { }
            map.UpdateProccess("Idle...");
        }

        private void BitmapEditor_Load(object sender, EventArgs e)
        {
            map.UpdateProccess("Loading Bitmap Raw...");
            BitmapRaw.Read(map);

            for (int i = 0; i < BitmapRaw.RawChunks.Count; i++)
            {
                listBox1.Items.Add("Bitmap Chunk #" + i.ToString());
            }
            try { listBox1.SelectedIndex = 0; }
            catch { }
            map.UpdateProccess("Idle...");
        }

        private void listBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            map.UpdateProccess("Loading Bitmap Raw...");
            int Index = listBox1.SelectedIndex;
            try { pictureBox1.Image.Dispose(); }
            catch { }
            pictureBox1.Image = Bitmap.LoadPreview(map, BitmapRaw.RawChunks[Index]);
            lblOffset.Text = BitmapRaw.RawChunks[Index].LOD1.ToString();
            lblSize.Text = BitmapRaw.RawChunks[Index].Size1.ToString();
            lblLocation.Text = BitmapRaw.RawChunks[Index].RawLocation.ToString();
            lblFormat.Text = BitmapRaw.RawChunks[Index].BitmapFormat.ToString();
            lblHeight.Text = BitmapRaw.RawChunks[Index].Height.ToString();
            lblWidth.Text = BitmapRaw.RawChunks[Index].Width.ToString();
            lblSwizzle.Text = BitmapRaw.RawChunks[Index].Swizzle.ToString();
            map.UpdateProccess("Idle...");
        }

        public void Reload()
        {
            BitmapRaw.Read(map);
            int Index = listBox1.SelectedIndex;
            pictureBox1.Image.Dispose();
            pictureBox1.Image = Bitmap.LoadPreview(map, BitmapRaw.RawChunks[Index]);
            lblOffset.Text = BitmapRaw.RawChunks[Index].LOD1.ToString();
            lblSize.Text = BitmapRaw.RawChunks[Index].Size1.ToString();
            lblLocation.Text = BitmapRaw.RawChunks[Index].RawLocation.ToString();
            lblFormat.Text = BitmapRaw.RawChunks[Index].BitmapFormat.ToString();
            lblHeight.Text = BitmapRaw.RawChunks[Index].Height.ToString();
            lblWidth.Text = BitmapRaw.RawChunks[Index].Width.ToString();
            lblSwizzle.Text = BitmapRaw.RawChunks[Index].Swizzle.ToString();
            map.UpdateProccess("Idle...");
        }

        private void btnExtract_Click(object sender, EventArgs e)
        {
            int Index = listBox1.SelectedIndex;
            SaveFileDialog sfd = new SaveFileDialog();
            if (sfd.ShowDialog() == DialogResult.OK)
            {
                map.UpdateProccess("Extracting...");
                DDS.Extract(map, BitmapRaw.RawChunks[Index], sfd.FileName);
                map.UpdateProccess("Idle...");
                MessageBox.Show("Done!");
            }
        }

        private void btnInject_Click(object sender, EventArgs e)
        {
            int Index = listBox1.SelectedIndex;
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.Filter = "DDS Images (*.dds)|*.dds";
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                map.UpdateProccess("Injecting...");
                DDS.Inject(map, BitmapRaw.RawChunks[Index], ofd.FileName);
                Reload();
                map.UpdateProccess("Idle...");
                MessageBox.Show("Done!");
            }
        }

        private void btnExtractAll_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog fbd = new FolderBrowserDialog();
            if (fbd.ShowDialog() == DialogResult.OK)
            {
                for (int i = 0; i < BitmapRaw.RawChunks.Count; i++)
                {
                    map.UpdateProccess("Extracting(" + i.ToString() + "/" + BitmapRaw.RawChunks.Count.ToString() + ")...");
                    DDS.Extract(map, BitmapRaw.RawChunks[i], fbd.SelectedPath + "\\Bitmap - " + i.ToString());
                }
                Reload();
                map.UpdateProccess("Idle...");
                MessageBox.Show("Done!");
            }
        }

        private void btnInternalize_Click(object sender, EventArgs e)
        {
            map.UpdateProccess("Internalizeing...");
            int Index = listBox1.SelectedIndex;
            DDS.Internalize(map, BitmapRaw.RawChunks[Index]);
            Reload();
            map.UpdateProccess("Idle...");
            MessageBox.Show("Done!");
        }

        private void radioButton1_CheckedChanged(object sender, EventArgs e)
        {
            pictureBox1.SizeMode = PictureBoxSizeMode.StretchImage;
        }

        private void radioButton2_CheckedChanged(object sender, EventArgs e)
        {
            pictureBox1.SizeMode = PictureBoxSizeMode.Normal;
        }
    }
}
