﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml;
using System.Windows.Forms;

namespace HaloMap
{
    public class PluginReader
    {
        public static XmlTextReader xr;
        public static List<Control> Controls = new List<Control>();
        public static List<string> Revisions = new List<string>();

        public static string Class;
        public static string Author;
        public static string Version;

        public static void Read(Map map, ref Panel Meta)
        {
            #region Reading
            Dispose();
            string C = map.SelectedTag.Class.Replace('*', '!');
            C = C.Replace('/', '_');
            C = C.Replace('\\', '_');
            C = C.Replace('<', '_');
            C = C.Replace('>', '_');
            string Location = Application.StartupPath + "\\Plugins\\" + C + ".Xml";

            if (!File.Exists(Location))
            {
                MessageBox.Show("Can not find plugin!");
            }
            else
            {
                xr = new XmlTextReader(new FileStream(Location, FileMode.Open, FileAccess.Read, FileShare.Read));

                Controls.Clear();
                Meta.Hide();
                Meta.Refresh();
                Meta.SuspendLayout();
                Class = "";
                Author = "";
                Version = "";
                while (xr.Read())
                {
                    switch (xr.NodeType)
                    {
                        case XmlNodeType.Element:
                            {
                                switch (xr.Name.ToLower())
                                {
                                    case "plugin":
                                        {
                                            Class = xr.GetAttribute("class");
                                            Author = xr.GetAttribute("author");
                                            Version = xr.GetAttribute("version");
                                            break;
                                        }
                                    case "revision":
                                        {
                                            Revisions.Add(xr.ReadElementString());
                                            break;
                                        }
                                    case "byte":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            Value b = new Value(Description, map.br.ReadByte(), typeof(byte), Offset);
                                            Controls.Add(b);
                                            b.Dock = DockStyle.Top;
                                            b.BringToFront();
                                            break;
                                        }
                                    case "short":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            Value b = new Value(Description, map.br.ReadInt16(), typeof(short), Offset);
                                            Controls.Add(b);
                                            b.Dock = DockStyle.Top;
                                            b.BringToFront();
                                            break;
                                        }
                                    case "ushort":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            Value b = new Value(Description, map.br.ReadUInt16(), typeof(ushort), Offset);
                                            Controls.Add(b);
                                            b.Dock = DockStyle.Top;
                                            b.BringToFront();
                                            break;
                                        }
                                    case "int":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            Value b = new Value(Description, map.br.ReadInt32(), typeof(int), Offset);
                                            Controls.Add(b);
                                            b.Dock = DockStyle.Top;
                                            b.BringToFront();
                                            break;
                                        }
                                    case "uint":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            Value b = new Value(Description, map.br.ReadUInt32(), typeof(uint), Offset);
                                            Controls.Add(b);
                                            b.Dock = DockStyle.Top;
                                            b.BringToFront();
                                            break;
                                        }
                                    case "float":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            Value b = new Value(Description, map.br.ReadSingle(), typeof(float), Offset);
                                            Controls.Add(b);
                                            b.Dock = DockStyle.Top;
                                            b.BringToFront();
                                            break;
                                        }
                                    case "string":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            int Length = Convert.ToInt32(xr.GetAttribute("length"));
                                            map.br.BaseStream.Position = Offset;
                                            Value b = new Value(Description, new string(map.br.ReadChars(Length)), typeof(char[]), Offset);
                                            Controls.Add(b);
                                            b.Dock = DockStyle.Top;
                                            b.BringToFront();
                                            break;
                                        }
                                    case "unknown":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            Value b = new Value(Description, map.br.ReadInt32(), typeof(int), Offset);
                                            Controls.Add(b);
                                            b.Dock = DockStyle.Top;
                                            b.BringToFront();
                                            break;
                                        }
                                    case "long":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            Value b = new Value(Description, map.br.ReadInt64(), typeof(long), Offset);
                                            Controls.Add(b);
                                            b.Dock = DockStyle.Top;
                                            b.BringToFront();
                                            break;
                                        }
                                    case "ident":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                            map.br.BaseStream.Position = Offset;
                                            TagReference t = new TagReference();
                                            t.Read(map, Offset);
                                            Ident i = new Ident(t, map, Description);
                                            Controls.Add(i);
                                            i.Dock = DockStyle.Top;
                                            i.BringToFront();
                                            break;
                                        }
                                    case "reflexive":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                            int Size = Convert.ToInt32(xr.GetAttribute("size"));
                                            map.br.BaseStream.Position = Offset;
                                            Reflexive r = new Reflexive();
                                            r.ChunkSize = Size;
                                            r.Read(map, Offset);
                                            Reflex rr = new Reflex(r, map, Description, xr.ReadSubtree(), Meta.Width - 50);
                                            Controls.Add(rr);
                                            rr.Dock = DockStyle.Top;
                                            rr.BringToFront();
                                            break;
                                        }
                                    case "stringid":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                            map.br.BaseStream.Position = Offset;
                                            ScriptReference s = new ScriptReference();
                                            s.Read(map, Offset);
                                            StringID sid = new StringID(s, map, Description);
                                            Controls.Add(sid);
                                            sid.Dock = DockStyle.Top;
                                            sid.BringToFront();
                                            break;
                                        }
                                    case "enum8":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            int Index = map.br.ReadByte();
                                            List<string> Items = new List<string>();
                                            List<int> Values = new List<int>();
                                            XmlReader x = xr.ReadSubtree();
                                            while (x.Read())
                                            {
                                                switch (x.Name.ToLower())
                                                {
                                                    case "option":
                                                        {
                                                            Items.Add(xr.GetAttribute("name"));
                                                            Values.Add(Convert.ToInt32(xr.GetAttribute("value")));
                                                            break;
                                                        }
                                                }
                                            }
                                            Enum e = new Enum(typeof(byte), Offset, Description, Items, Values, Index);
                                            Controls.Add(e);
                                            e.Dock = DockStyle.Top;
                                            e.BringToFront();
                                            break;
                                        }
                                    case "enum16":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            int Index = map.br.ReadInt16();
                                            List<string> Items = new List<string>();
                                            List<int> Values = new List<int>();
                                            XmlReader x = xr.ReadSubtree();
                                            while (x.Read())
                                            {
                                                switch (x.Name.ToLower())
                                                {
                                                    case "option":
                                                        {
                                                            Items.Add(xr.GetAttribute("name"));
                                                            Values.Add(Convert.ToInt32(xr.GetAttribute("value")));
                                                            break;
                                                        }
                                                }
                                            }
                                            Enum e = new Enum(typeof(short), Offset, Description, Items, Values, Index);
                                            Controls.Add(e);
                                            e.Dock = DockStyle.Top;
                                            e.BringToFront();
                                            break;
                                        }
                                    case "enum32":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            int Index = map.br.ReadInt32();
                                            List<string> Items = new List<string>();
                                            List<int> Values = new List<int>();
                                            XmlReader x = xr.ReadSubtree();
                                            while (x.Read())
                                            {
                                                switch (x.Name.ToLower())
                                                {
                                                    case "option":
                                                        {
                                                            Items.Add(xr.GetAttribute("name"));
                                                            Values.Add(Convert.ToInt32(xr.GetAttribute("value")));
                                                            break;
                                                        }
                                                }
                                            }
                                            Enum e = new Enum(typeof(int), Offset, Description, Items, Values, Index);
                                            Controls.Add(e);
                                            e.Dock = DockStyle.Top;
                                            e.BringToFront();
                                            break;
                                        }
                                    case "bitmask8":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            byte Index = map.br.ReadByte();
                                            List<string> Items = new List<string>();
                                            List<int> Values = new List<int>();
                                            XmlReader x = xr.ReadSubtree();
                                            while (x.Read())
                                            {
                                                switch (x.Name.ToLower())
                                                {
                                                    case "option":
                                                        {
                                                            Items.Add(xr.GetAttribute("name"));
                                                            Values.Add(Convert.ToInt32(xr.GetAttribute("value")));
                                                            break;
                                                        }
                                                }
                                            }
                                            Bitmask b = new Bitmask(Description, Index, Items, Values, Offset, typeof(byte));
                                            Controls.Add(b);
                                            b.Dock = DockStyle.Top;
                                            b.BringToFront();
                                            break;
                                        }
                                    case "bitmask16":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            ushort Index = map.br.ReadUInt16();
                                            List<string> Items = new List<string>();
                                            List<int> Values = new List<int>();
                                            XmlReader x = xr.ReadSubtree();
                                            while (x.Read())
                                            {
                                                switch (x.Name.ToLower())
                                                {
                                                    case "option":
                                                        {
                                                            Items.Add(xr.GetAttribute("name"));
                                                            Values.Add(Convert.ToInt32(xr.GetAttribute("value")));
                                                            break;
                                                        }
                                                }
                                            }
                                            Bitmask b = new Bitmask(Description, Index, Items, Values, Offset, typeof(short));
                                            Controls.Add(b);
                                            b.Dock = DockStyle.Top;
                                            b.BringToFront();
                                            break;
                                        }
                                    case "bitmask32":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            uint Index = map.br.ReadUInt32();
                                            List<string> Items = new List<string>();
                                            List<int> Values = new List<int>();
                                            XmlReader x = xr.ReadSubtree();
                                            while (x.Read())
                                            {
                                                switch (x.Name.ToLower())
                                                {
                                                    case "option":
                                                        {
                                                            Items.Add(xr.GetAttribute("name"));
                                                            Values.Add(Convert.ToInt32(xr.GetAttribute("value")));
                                                            break;
                                                        }
                                                }
                                            }
                                            Bitmask b = new Bitmask(Description, Index, Items, Values, Offset, typeof(int));
                                            Controls.Add(b);
                                            b.Dock = DockStyle.Top;
                                            b.BringToFront();
                                            break;
                                        }
                                    case "string32":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            string Value = new string(map.br.ReadChars(32));
                                            String s = new String(Description, Offset, Value, StringType.String32);
                                            Controls.Add(s);
                                            s.Dock = DockStyle.Top;
                                            s.BringToFront();
                                            break;
                                        }
                                    case "string64":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            string Value = new string(map.br.ReadChars(64));
                                            String s = new String(Description, Offset, Value, StringType.String64);
                                            Controls.Add(s);
                                            s.Dock = DockStyle.Top;
                                            s.BringToFront();
                                            break;
                                        }
                                    case "unicode64":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            string Value = "";
                                            for (int i = 0; i < 32; i++)
                                            {
                                                Value += map.br.ReadChar();
                                                map.br.BaseStream.Position++;
                                            }
                                            String s = new String(Description, Offset, Value, StringType.Unicode64);
                                            Controls.Add(s);
                                            s.Dock = DockStyle.Top;
                                            s.BringToFront();
                                            break;
                                        }
                                    case "unicode256":
                                        {
                                            string Description = xr.GetAttribute("name");
                                            int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + map.SelectedTag.MetaOffset;
                                            map.br.BaseStream.Position = Offset;
                                            string Value = "";
                                            for (int i = 0; i < 128; i++)
                                            {
                                                Value += map.br.ReadChar();
                                                map.br.BaseStream.Position++;
                                            }
                                            String s = new String(Description, Offset, Value, StringType.Unicode256);
                                            Controls.Add(s);
                                            s.Dock = DockStyle.Top;
                                            s.BringToFront();
                                            break;
                                        }
                                }
                                break;
                            }
                    }
                }

                xr.Close();

                //Load Controls
                for (int i = Controls.Count - 1; i > -1; i--)
                {
                    Meta.Controls.Add(Controls[i]);
                }
                Meta.ResumeLayout(true);
                Meta.Show();
            }
            #endregion
        }

        public static void Save(Map map, ref Panel Meta)
        {
            #region Saving
            for (int i = 0; i < Meta.Controls.Count; i++)
            {
                if (Meta.Controls[i].GetType() == typeof(Value))
                {
                    Value v = (Value)Meta.Controls[i];
                    map.bw.BaseStream.Position = v.Offset;
                    object Val = v.Save();
                    if (v.t == typeof(byte))
                    {
                        map.bw.Write(byte.Parse(Val.ToString()));
                    }
                    else if (v.t == typeof(short))
                    {
                        map.bw.Write(short.Parse(Val.ToString()));
                    }
                    else if (v.t == typeof(ushort))
                    {
                        map.bw.Write(ushort.Parse(Val.ToString()));
                    }
                    else if (v.t == typeof(int))
                    {
                        map.bw.Write(int.Parse(Val.ToString()));
                    }
                    else if (v.t == typeof(uint))
                    {
                        map.bw.Write(uint.Parse(Val.ToString()));
                    }
                    else if (v.t == typeof(float))
                    {
                        map.bw.Write(float.Parse(Val.ToString()));
                    }
                    else if (v.t == typeof(char[]))
                    {
                        string s = Val.ToString();
                        map.bw.Write(s.ToCharArray());
                    }
                }
                else if (Meta.Controls[i].GetType() == typeof(Ident))
                {
                    Ident id = (Ident)Meta.Controls[i];
                    TagReference t = id.Save();
                    t.Write(map);
                }
                else if (Meta.Controls[i].GetType() == typeof(Reflex))
                {
                    Reflex r = (Reflex)Meta.Controls[i];
                    r.Save();
                }
                else if (Meta.Controls[i].GetType() == typeof(StringID))
                {
                    StringID sid = (StringID)Meta.Controls[i];
                    ScriptReference s = sid.Save();
                    s.Write(map);
                }
                else if (Meta.Controls[i].GetType() == typeof(Enum))
                {
                    Enum e = (Enum)Meta.Controls[i];
                    map.bw.BaseStream.Position = e.Offset;
                    object Val = e.Save();
                    if (e.t == typeof(byte))
                    {
                        try { map.bw.Write(byte.Parse(Val.ToString())); }
                        catch (Exception) { map.bw.Write((byte)255); }
                    }
                    else if (e.t == typeof(short))
                    {
                        map.bw.Write(short.Parse(Val.ToString()));
                    }
                    else if (e.t == typeof(int))
                    {
                        map.bw.Write(int.Parse(Val.ToString()));
                    }
                }
                else if (Meta.Controls[i].GetType() == typeof(Bitmask))
                {
                    Bitmask b = (Bitmask)Meta.Controls[i];
                    map.bw.BaseStream.Position = b.Offset;
                    uint Value = b.Save();
                    object Val = Value;
                    if (b.t == typeof(byte))
                    {
                        map.bw.Write(byte.Parse(Val.ToString()));
                    }
                    else if (b.t == typeof(short))
                    {
                        map.bw.Write(ushort.Parse(Val.ToString()));
                    }
                    else if (b.t == typeof(int))
                    {
                        map.bw.Write(uint.Parse(Val.ToString()));
                    }
                }
                else if (Meta.Controls[i].GetType() == typeof(String))
                {
                    String s = (String)Meta.Controls[i];
                    map.bw.BaseStream.Position = s.Offset;
                    if (s.t == StringType.String32)
                    {
                        map.bw.Write(s.Save().ToCharArray());
                        for (int x = s.Value.Length; x < 32; x++)
                        {
                            map.bw.Write((byte)0);
                        }
                    }
                    else if (s.t == StringType.String64)
                    {
                        map.bw.Write(s.Save().ToCharArray());
                        for (int x = s.Value.Length; x < 64; x++)
                        {
                            map.bw.Write((byte)0);
                        }
                    }
                    else if (s.t == StringType.Unicode64)
                    {
                        for (int x = 0; x < 32; x++)
                        {
                            try { map.bw.Write(s.Value[x]); }
                            catch { map.bw.Write((byte)0); }
                            map.bw.Write((byte)0);
                        }
                    }
                    else if (s.t == StringType.Unicode256)
                    {
                        for (int x = 0; x < 128; x++)
                        {
                            try { map.bw.Write(s.Value[x]); }
                            catch { map.bw.Write((byte)0); }
                            map.bw.Write((byte)0);
                        }
                    }
                }
            }
            #endregion
        }

        public static void Poke(Map map, ref Panel Meta)
        {
            #region Poke
            if (RealTimeHalo.Initialize(map))
            {
                for (int i = 0; i < Meta.Controls.Count; i++)
                {
                    if (Meta.Controls[i].GetType() == typeof(Value))
                    {
                        Value v = (Value)Meta.Controls[i];
                        v.Poke(RealTimeHalo.ps, map.SelectedTag.MetaOffset);
                    }
                    else if (Meta.Controls[i].GetType() == typeof(Ident))
                    {
                        Ident id = (Ident)Meta.Controls[i];
                        id.Poke(RealTimeHalo.ps, map.SelectedTag.MetaOffset);
                    }
                    else if (Meta.Controls[i].GetType() == typeof(Reflex))
                    {
                        Reflex r = (Reflex)Meta.Controls[i];
                        r.Poke(RealTimeHalo.ps);
                    }
                    else if (Meta.Controls[i].GetType() == typeof(StringID))
                    {
                        StringID s = (StringID)Meta.Controls[i];
                        s.Poke(RealTimeHalo.ps, map.SelectedTag.MetaOffset);
                    }
                    else if (Meta.Controls[i].GetType() == typeof(Enum))
                    {
                        Enum e = (Enum)Meta.Controls[i];
                        e.Poke(RealTimeHalo.ps, map.SelectedTag.MetaOffset);
                    }
                    else if (Meta.Controls[i].GetType() == typeof(Bitmask))
                    {
                        Bitmask b = (Bitmask)Meta.Controls[i];
                        b.Poke(RealTimeHalo.ps, map.SelectedTag.MetaOffset);
                    }
                    else if (Meta.Controls[i].GetType() == typeof(String))
                    {
                        String s = (String)Meta.Controls[i];
                        s.Poke(RealTimeHalo.ps, map.SelectedTag.MetaOffset);
                    }
                }
            }
            #endregion
        }

        public static void Dispose()
        {
            #region Disposing
            foreach (Control c in Controls)
            {
                c.Dispose();
            }
            Controls.Clear();
            Class = null;
            Author = null;
            Version = null;
            GC.Collect();
            #endregion
        }

        #region References
        public static void ReadReferences(Map map)
        {
            //map.UpdateProccess("Reading Idents...");
            LoadIdents(map);
            //map.UpdateProccess("Reading Reflexives...");
            LoadReflexives(map);
            //map.UpdateProccess("Reading Script References...");
            LoadScriptReferences(map);
        }

        public static void LoadIdents(Map map)
        {
            string C = map.SelectedTag.Class.Replace('*', '!');
            C = C.Replace('/', '_');
            C = C.Replace('\\', '_');
            C = C.Replace('<', '_');
            C = C.Replace('>', '_');
            string Location = Application.StartupPath + "\\Plugins\\" + C + ".Xml";

            if (!File.Exists(Location))
            {
                MessageBox.Show("Can not find plugin!");
                return;
            }
            else
            {
                xr = new XmlTextReader(new FileStream(Location, FileMode.Open, FileAccess.Read, FileShare.Read));

                map.SelectedTag.Idents.Clear();

                //Read
                while (xr.Read())
                {
                    switch (xr.NodeType)
                    {
                        case XmlNodeType.Element:
                            {
                                switch (xr.Name.ToLower())
                                {
                                    case "plugin":
                                        {
                                            Author = xr.GetAttribute("author");
                                            Version = xr.GetAttribute("version");
                                            break;
                                        }
                                    case "ident":
                                        {
                                            TagReference t = new TagReference();
                                            t.Description = xr.GetAttribute("name");
                                            t.Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                            t.Read(map);
                                            map.SelectedTag.Idents.Add(t);
                                            break;
                                        }
                                    case "reflexive":
                                        {
                                            Reflexive r = new Reflexive();
                                            r.Description = xr.GetAttribute("name");
                                            r.Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                            r.ChunkSize = Convert.ToInt32(xr.GetAttribute("size"));
                                            r.Read(map);
                                            if (r.ChunkCount > 0)
                                            {
                                                RecrusivlyLoadIdents(map, r, xr.ReadSubtree());
                                            }
                                            else
                                            {
                                                xr.Skip();
                                            }
                                            break;
                                        }
                                }
                                break;
                            }
                    }
                }
                xr.Close();
            }
        }

        public static void LoadReflexives(Map map)
        {
            string C = map.SelectedTag.Class.Replace('*', '!');
            C = C.Replace('/', '_');
            C = C.Replace('\\', '_');
            C = C.Replace('<', '_');
            C = C.Replace('>', '_');
            string Location = Application.StartupPath + "\\Plugins\\" + C + ".Xml";

            if (!File.Exists(Location))
            {
                MessageBox.Show("Can not find plugin!");
                return;
            }
            else
            {
                xr = new XmlTextReader(new FileStream(Location, FileMode.Open, FileAccess.Read, FileShare.Read));

                map.SelectedTag.Reflexives.Clear();

                //Read
                while (xr.Read())
                {
                    switch (xr.NodeType)
                    {
                        case XmlNodeType.Element:
                            {
                                switch (xr.Name.ToLower())
                                {
                                    case "plugin":
                                        {
                                            Author = xr.GetAttribute("author");
                                            Version = xr.GetAttribute("version");
                                            break;
                                        }
                                    case "reflexive":
                                        {
                                            Reflexive r = new Reflexive();
                                            r.Description = xr.GetAttribute("name");
                                            r.Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                            r.ChunkSize = Convert.ToInt32(xr.GetAttribute("size"));
                                            r.Read(map);
                                            map.SelectedTag.Reflexives.Add(r);
                                            if (r.ChunkCount > 0)
                                            {
                                                RecrusivlyLoadReflexives(map, r, xr.ReadSubtree());
                                            }
                                            else
                                            {
                                                xr.Skip();
                                            }
                                            break;
                                        }
                                }
                                break;
                            }
                    }
                }
                xr.Close();
            }
        }

        public static void LoadScriptReferences(Map map)
        {
            string C = map.SelectedTag.Class.Replace('*', '!');
            C = C.Replace('/', '_');
            C = C.Replace('\\', '_');
            C = C.Replace('<', '_');
            C = C.Replace('>', '_');
            string Location = Application.StartupPath + "\\Plugins\\" + C + ".Xml";

            if (!File.Exists(Location))
            {
                MessageBox.Show("Can not find plugin!");
                return;
            }
            else
            {
                xr = new XmlTextReader(new FileStream(Location, FileMode.Open, FileAccess.Read, FileShare.Read));

                map.SelectedTag.ScriptReferences.Clear();

                //Read
                while (xr.Read())
                {
                    switch (xr.NodeType)
                    {
                        case XmlNodeType.Element:
                            {
                                switch (xr.Name.ToLower())
                                {
                                    case "plugin":
                                        {
                                            Author = xr.GetAttribute("author");
                                            Version = xr.GetAttribute("version");
                                            break;
                                        }
                                    case "stringid":
                                        {
                                            ScriptReference s = new ScriptReference();
                                            s.Description = xr.GetAttribute("name");
                                            s.Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                            s.Read(map);
                                            map.SelectedTag.ScriptReferences.Add(s);
                                            break;
                                        }
                                    case "reflexive":
                                        {
                                            Reflexive r = new Reflexive();
                                            r.Description = xr.GetAttribute("name");
                                            r.Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                            r.ChunkSize = Convert.ToInt32(xr.GetAttribute("size"));
                                            r.Read(map);
                                            if (r.ChunkCount > 0)
                                            {
                                                RecrusivlyLoadScripts(map, r, xr.ReadSubtree());
                                            }
                                            else
                                            {
                                                xr.Skip();
                                            }
                                            break;
                                        }
                                }
                                break;
                            }
                    }
                }
                xr.Close();
            }
        }
        #endregion

        #region Recrusivly Load References
        private static void RecrusivlyLoadIdents(Map map, Reflexive r, XmlReader xr)
        {
            xr.ReadStartElement();
            while (xr.Read())
            {
                switch (xr.NodeType)
                {
                    case XmlNodeType.Element:
                        {
                            switch (xr.Name.ToLower())
                            {
                                case "ident":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + (r.Translation - map.SelectedTag.MetaOffset);
                                        for (int i = 0; i < r.ChunkCount; i++)
                                        {
                                            TagReference t = new TagReference();
                                            t.Description = Description + "[" + i.ToString() + "]";
                                            t.Offset = Offset + (r.ChunkSize * i);
                                            t.Read(map, t.Offset);
                                            map.SelectedTag.Idents.Add(t);
                                        }
                                        break;
                                    }
                                case "reflexive":
                                    {
                                        Reflexive rr = new Reflexive();
                                        rr.Description = xr.GetAttribute("name");
                                        rr.Offset = Convert.ToInt32(xr.GetAttribute("offset")) + r.Translation;
                                        rr.ChunkSize = Convert.ToInt32(xr.GetAttribute("size"));
                                        rr.Read(map, rr.Offset, true);
                                        if (rr.ChunkCount > 0)
                                        {
                                            RecrusivlyLoadIdents(map, rr, xr.ReadSubtree());
                                        }
                                        else
                                        {
                                            xr.Skip();
                                        }
                                        break;
                                    }
                            }
                            break;
                        }
                }
            }
        }
        private static void RecrusivlyLoadScripts(Map map, Reflexive r, XmlReader xr)
        {
            xr.ReadStartElement();
            while (xr.Read())
            {
                switch (xr.NodeType)
                {
                    case XmlNodeType.Element:
                        {
                            switch (xr.Name.ToLower())
                            {
                                case "stringid":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + (r.Translation - map.SelectedTag.MetaOffset);
                                        for (int i = 0; i < r.ChunkCount; i++)
                                        {
                                            ScriptReference s = new ScriptReference();
                                            s.Description = Description + "[" + i.ToString() + "]";
                                            s.Offset = Offset + (r.ChunkSize * i);
                                            s.Read(map, s.Offset);
                                            map.SelectedTag.ScriptReferences.Add(s);
                                        }
                                        break;
                                    }
                                case "reflexive":
                                    {
                                        Reflexive rr = new Reflexive();
                                        rr.Description = xr.GetAttribute("name");
                                        rr.Offset = Convert.ToInt32(xr.GetAttribute("offset")) + r.Translation;
                                        rr.ChunkSize = Convert.ToInt32(xr.GetAttribute("size"));
                                        rr.Read(map, rr.Offset, true);
                                        if (rr.ChunkCount > 0)
                                        {
                                            RecrusivlyLoadScripts(map, rr, xr.ReadSubtree());
                                        }
                                        else
                                        {
                                            xr.Skip();
                                        }
                                        break;
                                    }
                            }
                            break;
                        }
                }
            }
        }
        private static void RecrusivlyLoadReflexives(Map map, Reflexive r, XmlReader xr)
        {
            xr.ReadStartElement();
            while (xr.Read())
            {
                switch (xr.NodeType)
                {
                    case XmlNodeType.Element:
                        {
                            switch (xr.Name.ToLower())
                            {
                                case "reflexive":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        int Size = Convert.ToInt32(xr.GetAttribute("size"));
                                        for (int i = 0; i < r.ChunkCount; i++)
                                        {
                                            if (r.RawTranslation != -1)
                                            {
                                                Reflexive rr = new Reflexive();
                                                rr.Description = Description + "[" + i.ToString() + "]";
                                                rr.Offset = Offset + (r.Translation - map.SelectedTag.MetaOffset) + (r.ChunkSize * i);
                                                rr.ChunkSize = Size;
                                                rr.Read(map, rr.Offset);
                                                map.SelectedTag.Reflexives.Add(rr);
                                                if (xr.NodeType == XmlNodeType.Element)
                                                {
                                                    RecrusivlyLoadReflexives(map, rr, xr.ReadSubtree());
                                                }
                                            }
                                        }
                                        break;
                                    }
                            }
                            break;
                        }
                    case XmlNodeType.EndElement:
                        {
                            break;
                        }
                }
            }
        }
        #endregion

        #region Parsing
        public static void ParseReflexives(Map map, int Shift)
        {
            string C = map.SelectedTag.Class.Replace('*', '!');
            C = C.Replace('/', '_');
            C = C.Replace('\\', '_');
            C = C.Replace('<', '_');
            C = C.Replace('>', '_');
            string Location = Application.StartupPath + "\\Plugins\\" + C + ".Xml";

            if (!File.Exists(Location))
            {
                MessageBox.Show("Can not find plugin!");
                return;
            }
            else
            {
                xr = new XmlTextReader(new FileStream(Location, FileMode.Open, FileAccess.Read, FileShare.Read));

                map.SelectedTag.Reflexives.Clear();

                //Read
                while (xr.Read())
                {
                    switch (xr.NodeType)
                    {
                        case XmlNodeType.Element:
                            {
                                switch (xr.Name.ToLower())
                                {
                                    case "plugin":
                                        {
                                            Author = xr.GetAttribute("author");
                                            Version = xr.GetAttribute("version");
                                            break;
                                        }
                                    case "reflexive":
                                        {
                                            Reflexive r = new Reflexive();
                                            r.Description = xr.GetAttribute("name");
                                            r.Offset = Convert.ToInt32(xr.GetAttribute("offset")) + Shift;
                                            r.ChunkSize = Convert.ToInt32(xr.GetAttribute("size"));
                                            r.Read(map);
                                            map.SelectedTag.Reflexives.Add(r);
                                            if (r.ChunkCount > 0)
                                            {
                                                RecrusivlyParseReflexives(map, r, xr.ReadSubtree(), Shift);
                                            }
                                            else
                                            {
                                                xr.Skip();
                                            }
                                            break;
                                        }
                                }
                                break;
                            }
                    }
                }
                xr.Close();
            }
        }

        private static void RecrusivlyParseReflexives(Map map, Reflexive r, XmlReader xr, int Shift)
        {
            xr.ReadStartElement();
            while (xr.Read())
            {
                switch (xr.NodeType)
                {
                    case XmlNodeType.Element:
                        {
                            switch (xr.Name.ToLower())
                            {
                                case "reflexive":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset")) +Shift;
                                        int Size = Convert.ToInt32(xr.GetAttribute("size"));
                                        for (int i = 0; i < r.ChunkCount; i++)
                                        {
                                            if (r.RawTranslation != -1)
                                            {
                                                Reflexive rr = new Reflexive();
                                                rr.Description = Description + "[" + i.ToString() + "]";
                                                rr.Offset = Offset + (r.Translation - map.SelectedTag.MetaOffset) + (r.ChunkSize * i); // +Shift;
                                                rr.ChunkSize = Size;
                                                rr.Read(map, rr.Offset);
                                                map.SelectedTag.Reflexives.Add(rr);
                                                if (xr.NodeType == XmlNodeType.Element)
                                                {
                                                    RecrusivlyParseReflexives(map, rr, xr.ReadSubtree(), Shift);
                                                }
                                            }
                                        }
                                        break;
                                    }
                            }
                            break;
                        }
                    case XmlNodeType.EndElement:
                        {
                            break;
                        }
                }
            }
        }
        #endregion
    }
}
