﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml;
using System.Windows.Forms;

namespace HaloMap
{
    public enum RawType
    {
        Sound,
        Model,
        BSP,
        Lightmap,
        Decorator,
        Particle,
        Weather,
        Animation,
        Bitmap,
        Coconut
    }

    public enum BitmapFormat
    {
        A1R5G5B5 = 8,
        A4R4G4B4 = 9,
        A8 = 0,
        A8R8G8B8 = 11,
        A8Y8 = 3,
        ARGBFP32 = 0x13,
        AY8 = 2,
        DXT1 = 14,
        DXT3 = 15,
        DXT5 = 0x10,
        G8B8 = 0x17,
        P8 = 0x12,
        P8Bump = 0x11,
        R5G6B5 = 6,
        RGBFP16 = 0x15,
        RGBFP32 = 20,
        V8U8 = 0x16,
        X8R8G8B8 = 10,
        Y8 = 1,
        Lightmap = 18,
        Unknown = 22
    }

    public class RawData
    {
        #region Methods
        public static RawLocation RawLocator(ref int Offset)
        {
            long Map = Offset & 0xC0000000;
            Offset &= 0x3FFFFFFF;
            switch (Map)
            {
                case 0:
                    {
                        return RawLocation.Internal;
                    }
                case 0x80000000:
                    {
                        return RawLocation.Shared;
                    }
                case 0xC0000000:
                    {
                        return RawLocation.SingleplayerShared;
                    }
                case 0x40000000:
                    {
                        return RawLocation.Mainmenu;
                    }
            }
            return RawLocation.Unknown;
        }
        public static RawLocation RawLocator(int Offset)
        {
            long Map = Offset & 0xC0000000;
            switch (Map)
            {
                case 0:
                    {
                        return RawLocation.Internal;
                    }
                case 0x80000000:
                    {
                        return RawLocation.Shared;
                    }
                case 0xC0000000:
                    {
                        return RawLocation.SingleplayerShared;
                    }
                case 0x40000000:
                    {
                        return RawLocation.Mainmenu;
                    }
            }
            return RawLocation.Unknown;
        }
        #endregion
    }

    public class RawChunk
    {
        public RawLocation RawLocation;
        public int Offset;
        public int Size;
        public byte[] Raw;
    }

    public class BitmapChunk
    {
        public RawLocation RawLocation;
        public BitmapFormat BitmapFormat;
        public int LOD1, LOD2, LOD3;
        public int Size1, Size2, Size3;
        public char[] Class;
        public ushort Width, Height, Depth, BitmapType, Flags, RegPointX, RegPointY, MipMapCount, PixelOffset;
        public bool Swizzle;
        public byte[] SourceData;
    }

    public static class BitmapRaw
    {
        public static List<BitmapChunk> RawChunks = new List<BitmapChunk>();

        #region Methods
        public static void Read(Map map)
        {
            RawChunks.Clear();
            //Goto the Bitmap Raw Reflexive
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 68;
            Reflexive r = new Reflexive();
            r.Read(map, true);
            //Read each Bitmap
            for (int i = 0; i < r.ChunkCount; i++)
            {
                BitmapChunk Bitm = new BitmapChunk();
                map.br.BaseStream.Position = r.Translation + (116 * i);
                //Class
                Bitm.Class = map.br.ReadChars(4);
                //Width
                Bitm.Width = map.br.ReadUInt16();
                //Height
                Bitm.Height = map.br.ReadUInt16();
                //Depth
                Bitm.Depth = map.br.ReadUInt16();
                //Type
                Bitm.BitmapType = map.br.ReadUInt16();
                //Format
                Bitm.BitmapFormat = (BitmapFormat)map.br.ReadUInt16();
                //Flags
                Bitm.Flags = map.br.ReadUInt16();
                if ((Bitm.Flags & 0x8) == 0x8)
                    Bitm.Swizzle = true;
                //RegPointX,Y
                Bitm.RegPointX = map.br.ReadUInt16();
                Bitm.RegPointY = map.br.ReadUInt16();
                //MipMapCount
                Bitm.MipMapCount = map.br.ReadUInt16();
                //Pixel Offset
                Bitm.PixelOffset = map.br.ReadUInt16();
                map.br.BaseStream.Position += 4;
                //LODs
                Bitm.LOD1 = map.br.ReadInt32();
                Bitm.LOD2 = map.br.ReadInt32();
                Bitm.LOD3 = map.br.ReadInt32();
                map.br.BaseStream.Position += 12;
                //Size
                Bitm.Size1 = map.br.ReadInt32();
                Bitm.Size2 = map.br.ReadInt32();
                Bitm.Size3 = map.br.ReadInt32();
                //RawLocation
                Bitm.RawLocation = RawData.RawLocator(ref Bitm.LOD1);//=D
                if (Bitm.LOD1 != -1 && Bitm.RawLocation != RawLocation.Internal)
                {
                    BinaryReader br = map.OpenMap(Bitm.RawLocation);//told u, ...lol
                    br.BaseStream.Position = Bitm.LOD1;//you havent modifed the offset, there pure offsets watch notice the offset is negative? well internal bitmaps are pure
                    Bitm.SourceData = br.ReadBytes(Bitm.Size1);
                    br.Close();
                }
                else if (Bitm.LOD1 != -1 && Bitm.RawLocation == RawLocation.Internal)
                {
                    map.br.BaseStream.Position = Bitm.LOD1;
                    Bitm.SourceData = map.br.ReadBytes(Bitm.Size1);
                }
                RawChunks.Add(Bitm);
            }
        }

        public static void Decompile(Map map, Tag tag, string Location)
        {
            XmlWriter xr = XmlWriter.Create(Location + ".Rawxml");
            xr.WriteStartDocument();
            RawChunks.Clear();
            //Goto the Bitmap Raw Reflexive
            map.br.BaseStream.Position = tag.MetaOffset + 68;
            Reflexive r = new Reflexive();
            r.Read(map, true);
            xr.WriteStartElement("Raw");
            xr.WriteElementString("RawChunks", r.ChunkCount.ToString());
            //Read each Bitmap
            for (int i = 0; i < r.ChunkCount; i++)
            {
                BitmapChunk Bitm = new BitmapChunk();
                xr.WriteStartElement("Chunk");
                map.br.BaseStream.Position = r.Translation + (116 * i);
                xr.WriteElementString("Offset", ((r.Translation + (116 * i)) - tag.MetaOffset).ToString());
                //Class
                Bitm.Class = map.br.ReadChars(4);
                xr.WriteElementString("Class", Bitm.Class.ToString());
                //Width
                Bitm.Width = map.br.ReadUInt16();
                xr.WriteElementString("Width", Bitm.Width.ToString());
                //Height
                Bitm.Height = map.br.ReadUInt16();
                xr.WriteElementString("Height", Bitm.Height.ToString());
                //Depth
                Bitm.Depth = map.br.ReadUInt16();
                xr.WriteElementString("Depth", Bitm.Depth.ToString());
                //Type
                Bitm.BitmapType = map.br.ReadUInt16();
                xr.WriteElementString("Type", Bitm.BitmapType.ToString());
                //Format
                Bitm.BitmapFormat = (BitmapFormat)map.br.ReadUInt16();
                xr.WriteElementString("Format", Bitm.BitmapFormat.ToString());
                //Flags
                Bitm.Flags = map.br.ReadUInt16();
                xr.WriteElementString("Flags", Bitm.Flags.ToString());
                if ((Bitm.Flags & 0x8) == 0x8)
                    Bitm.Swizzle = true;
                xr.WriteElementString("Swizzel", Bitm.Swizzle.ToString());
                //RegPointX,Y
                Bitm.RegPointX = map.br.ReadUInt16();
                xr.WriteElementString("RegPointX", Bitm.RegPointX.ToString());
                Bitm.RegPointY = map.br.ReadUInt16();
                xr.WriteElementString("RegPointY", Bitm.RegPointY.ToString());
                //MipMapCount
                Bitm.MipMapCount = map.br.ReadUInt16();
                xr.WriteElementString("MipMapCount", Bitm.MipMapCount.ToString());
                //Pixel Offset
                Bitm.PixelOffset = map.br.ReadUInt16();
                xr.WriteElementString("PixelOffset", Bitm.PixelOffset.ToString());
                map.br.BaseStream.Position += 4;
                //LODs
                Bitm.LOD1 = map.br.ReadInt32();
                xr.WriteElementString("LOD1", Bitm.LOD1.ToString());
                Bitm.LOD2 = map.br.ReadInt32();
                xr.WriteElementString("LOD2", Bitm.LOD2.ToString());
                Bitm.LOD3 = map.br.ReadInt32();
                xr.WriteElementString("LOD3", Bitm.LOD3.ToString());
                map.br.BaseStream.Position += 12;
                //Size
                Bitm.Size1 = map.br.ReadInt32();
                xr.WriteElementString("Size1", Bitm.Size1.ToString());
                Bitm.Size2 = map.br.ReadInt32();
                xr.WriteElementString("Size2", Bitm.Size2.ToString());
                Bitm.Size3 = map.br.ReadInt32();
                xr.WriteElementString("Size3", Bitm.Size3.ToString());
                //RawLocation
                Bitm.RawLocation = RawData.RawLocator(ref Bitm.LOD1);
                xr.WriteElementString("Location", Bitm.RawLocation.ToString());
                if (Bitm.LOD1 != -1 && Bitm.RawLocation != RawLocation.Internal)
                {
                    BinaryReader br = map.OpenMap(Bitm.RawLocation);
                    br.BaseStream.Position = Bitm.LOD1;
                    Bitm.SourceData = br.ReadBytes(Bitm.Size1);
                    br.Close();
                }
                else if (Bitm.LOD1 != -1 && Bitm.RawLocation == RawLocation.Internal)
                {
                    map.br.BaseStream.Position = Bitm.LOD1;
                    Bitm.SourceData = map.br.ReadBytes(Bitm.Size1);
                }
                xr.WriteEndElement();
                RawChunks.Add(Bitm);
            }
            xr.WriteEndElement();
            xr.Close();

            //Write Raw
            for (int i = 0; i < RawChunks.Count; i++)
            {
                //Write Raw
                BinaryWriter bw = new BinaryWriter(new FileStream(Location + ".Raw", FileMode.OpenOrCreate, FileAccess.Write, FileShare.Write));
                bw.BaseStream.Position = bw.BaseStream.Length;
                if (RawChunks[i].SourceData != null)
                {
                    bw.Write(RawChunks[i].SourceData);
                }
                bw.Close();
            }
        }
        #endregion
    }

    public static class ModelRaw
    {
        public static List<RawChunk> Model1 = new List<RawChunk>();
        public static List<RawChunk> Model2 = new List<RawChunk>();

        #region Methods
        public static void Read(Map map)
        {
            Model1.Clear();
            Model2.Clear();

            //Model-1
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 36;
            Reflexive r = new Reflexive();
            r.Read(map, true);
            for (int i = 0; i < r.ChunkCount; i++)
            {
                RawChunk Model = new RawChunk();
                map.br.BaseStream.Position = r.Translation + (i * 92) + 56;
                Model.Offset = map.br.ReadInt32();
                Model.Size = map.br.ReadInt32();
                Model.RawLocation = RawData.RawLocator(ref Model.Offset);
                if (Model.RawLocation == RawLocation.Internal && Model.Offset != -1)
                {
                    map.br.BaseStream.Position = Model.Offset;
                    Model.Raw = map.br.ReadBytes(Model.Size);
                }
                else if (Model.RawLocation != RawLocation.Internal && Model.Offset != -1)
                {
                    BinaryReader br = map.OpenMap(Model.RawLocation);
                    br.BaseStream.Position = Model.Offset;
                    Model.Raw = br.ReadBytes(Model.Size);
                    br.Close();
                }
                Model1.Add(Model);
            }

            //Model-2
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 116;
            r.Read(map, true);
            for (int i = 0; i < r.ChunkCount; i++)
            {
                RawChunk Model = new RawChunk();
                map.br.BaseStream.Position = r.Translation + (i * 88) + 52;
                Model.Offset = map.br.ReadInt32();
                Model.Size = map.br.ReadInt32();
                Model.RawLocation = RawData.RawLocator(ref Model.Offset);
                if (Model.RawLocation == RawLocation.Internal && Model.Offset != -1)
                {
                    map.br.BaseStream.Position = Model.Offset;
                    Model.Raw = map.br.ReadBytes(Model.Size);
                }
                else if (Model.RawLocation != RawLocation.Internal && Model.Offset != -1)
                {
                    BinaryReader br = map.OpenMap(Model.RawLocation);
                    br.BaseStream.Position = Model.Offset;
                    Model.Raw = br.ReadBytes(Model.Size);
                    br.Close();
                }
                Model2.Add(Model);
            }
        }
        public static void Decompile(Map map, Tag tag, string Location)
        {
            Model1.Clear();
            Model2.Clear();

            //Set up Xml
            XmlWriter xr = XmlWriter.Create(Location + ".Rawxml");
            xr.WriteStartDocument();
            xr.WriteStartElement("Raw");

            //Model-1
            map.br.BaseStream.Position = tag.MetaOffset + 36;
            Reflexive r = new Reflexive();
            r.Read(map, true);
            xr.WriteStartElement("Model1");
            xr.WriteElementString("RawChunks", r.ChunkCount.ToString());
            for (int i = 0; i < r.ChunkCount; i++)
            {
                RawChunk Model = new RawChunk();
                xr.WriteStartElement("Chunk");
                map.br.BaseStream.Position = r.Translation + (i * 92) + 56;
                Model.Offset = map.br.ReadInt32();
                xr.WriteElementString("Offset", Model.Offset.ToString());
                Model.Size = map.br.ReadInt32();
                xr.WriteElementString("Size", Model.Size.ToString());
                Model.RawLocation = RawData.RawLocator(ref Model.Offset);
                xr.WriteElementString("Location", Model.RawLocation.ToString());
                if (Model.RawLocation == RawLocation.Internal && Model.Offset != -1)
                {
                    map.br.BaseStream.Position = Model.Offset;
                    Model.Raw = map.br.ReadBytes(Model.Size);
                }
                else if (Model.RawLocation != RawLocation.Internal && Model.Offset != -1)
                {
                    BinaryReader br = map.OpenMap(Model.RawLocation);
                    br.BaseStream.Position = Model.Offset;
                    Model.Raw = br.ReadBytes(Model.Size);
                    br.Close();
                }
                Model1.Add(Model);
                xr.WriteEndElement();
            }
            xr.WriteEndElement();

            //Write Raw
            for (int i = 0; i < Model1.Count; i++)
            {
                //Write Raw
                BinaryWriter bw = new BinaryWriter(new FileStream(Location + ".Model1", FileMode.OpenOrCreate, FileAccess.Write, FileShare.Write));
                bw.BaseStream.Position = bw.BaseStream.Length;
                if (Model1[i].Raw != null)
                {
                    bw.Write(Model1[i].Raw);
                }
                bw.Close();
            }


            //Model-2
            map.br.BaseStream.Position = tag.MetaOffset + 116;
            r.Read(map, true);
            xr.WriteStartElement("Model2");
            xr.WriteElementString("RawChunks", r.ChunkCount.ToString());
            for (int i = 0; i < r.ChunkCount; i++)
            {
                RawChunk Model = new RawChunk();
                xr.WriteStartElement("Chunk");
                map.br.BaseStream.Position = r.Translation + (i * 88) + 52;
                Model.Offset = map.br.ReadInt32();
                xr.WriteElementString("Offset", Model.Offset.ToString());
                Model.Size = map.br.ReadInt32();
                xr.WriteElementString("Size", Model.Size.ToString());
                Model.RawLocation = RawData.RawLocator(ref Model.Offset);
                xr.WriteElementString("Location", Model.RawLocation.ToString());
                if (Model.RawLocation == RawLocation.Internal && Model.Offset != -1)
                {
                    map.br.BaseStream.Position = Model.Offset;
                    Model.Raw = map.br.ReadBytes(Model.Size);
                }
                else if (Model.RawLocation != RawLocation.Internal && Model.Offset != -1)
                {
                    BinaryReader br = map.OpenMap(Model.RawLocation);
                    br.BaseStream.Position = Model.Offset;
                    Model.Raw = br.ReadBytes(Model.Size);
                    br.Close();
                }
                Model2.Add(Model);
                xr.WriteEndElement();
            }
            xr.WriteEndElement();
            xr.WriteEndElement();
            xr.Close();

            //Write Raw
            for (int i = 0; i < Model2.Count; i++)
            {
                //Write Raw
                BinaryWriter bw = new BinaryWriter(new FileStream(Location + ".Model2", FileMode.OpenOrCreate, FileAccess.Write, FileShare.Write));
                bw.BaseStream.Position = bw.BaseStream.Length;
                if (Model2[i].Raw != null)
                {
                    bw.Write(Model2[i].Raw);
                }
                bw.Close();
            }
        }
        #endregion
    }

    public static class DecoratorRaw
    {
        public static List<RawChunk> Decorator = new List<RawChunk>();

        #region Methods
        public static void Read(Map map)
        {
            Decorator.Clear();

            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 56;
            RawChunk Deco = new RawChunk();
            Deco.Offset = map.br.ReadInt32();
            Deco.Size = map.br.ReadInt32();
            Deco.RawLocation = RawData.RawLocator(ref Deco.Offset);
            if (Deco.RawLocation == RawLocation.Internal && Deco.Offset != -1)
            {
                map.br.BaseStream.Position = Deco.Offset;
                Deco.Raw = map.br.ReadBytes(Deco.Size);
            }
            else if (Deco.RawLocation != RawLocation.Internal && Deco.Offset != -1)
            {
                BinaryReader br = map.OpenMap(Deco.RawLocation);
                br.BaseStream.Position = Deco.Offset;
                Deco.Raw = br.ReadBytes(Deco.Size);
                br.Close();
            }
            Decorator.Add(Deco);
        }

        public static void Decompile(Map map, Tag tag, string Location)
        {
            Decorator.Clear();

            //Set up Xml
            XmlWriter xr = XmlWriter.Create(Location + ".Rawxml");
            xr.WriteStartDocument();

            map.br.BaseStream.Position = tag.MetaOffset + 56;
            xr.WriteStartElement("Raw");
            RawChunk Deco = new RawChunk();
            Deco.Offset = map.br.ReadInt32();
            xr.WriteElementString("Offset", Deco.Offset.ToString());
            Deco.Size = map.br.ReadInt32();
            xr.WriteElementString("Size", Deco.Size.ToString());
            Deco.RawLocation = RawData.RawLocator(ref Deco.Offset);
            xr.WriteElementString("Location", Deco.RawLocation.ToString());
            if (Deco.RawLocation == RawLocation.Internal && Deco.Offset != -1)
            {
                map.br.BaseStream.Position = Deco.Offset;
                Deco.Raw = map.br.ReadBytes(Deco.Size);
            }
            else if (Deco.RawLocation != RawLocation.Internal && Deco.Offset != -1)
            {
                BinaryReader br = map.OpenMap(Deco.RawLocation);
                br.BaseStream.Position = Deco.Offset;
                Deco.Raw = br.ReadBytes(Deco.Size);
                br.Close();
            }
            Decorator.Add(Deco);
            xr.WriteEndElement();
            xr.Close();

            //Write Raw
            if (Deco.Raw != null)
            {
                File.WriteAllBytes(Location + ".Raw", Deco.Raw);
            }
        }
        #endregion
    }

    public static class ParticleRaw
    {
        public static List<RawChunk> Particles = new List<RawChunk>();

        #region Methods
        public static void Read(Map map)
        {
            Particles.Clear();
            RawChunk Part = new RawChunk();
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 160;
            Part.Offset = map.br.ReadInt32();
            Part.Size = map.br.ReadInt32();
            Part.RawLocation = RawData.RawLocator(ref Part.Offset);
            if (Part.RawLocation == RawLocation.Internal && Part.Offset != -1)
            {
                map.br.BaseStream.Position = Part.Offset;
                Part.Raw = map.br.ReadBytes(Part.Size);
            }
            else if (Part.RawLocation != RawLocation.Internal && Part.Offset != -1)
            {
                BinaryReader br = map.OpenMap(Part.RawLocation);
                br.BaseStream.Position = Part.Offset;
                Part.Raw = br.ReadBytes(Part.Size);
                br.Close();
            }
            Particles.Add(Part);
        }

        public static void Decompile(Map map, Tag tag, string Location)
        {
            Particles.Clear();

            //Set up Xml
            XmlWriter xr = XmlWriter.Create(Location + ".Rawxml");
            xr.WriteStartDocument();
            xr.WriteStartElement("Raw");

            RawChunk Part = new RawChunk();
            map.br.BaseStream.Position = tag.MetaOffset + 160;
            Part.Offset = map.br.ReadInt32();
            xr.WriteElementString("Offset", Part.Offset.ToString());
            Part.Size = map.br.ReadInt32();
            xr.WriteElementString("Size", Part.Size.ToString());
            Part.RawLocation = RawData.RawLocator(ref Part.Offset);
            xr.WriteElementString("Location", Part.RawLocation.ToString());
            if (Part.RawLocation == RawLocation.Internal && Part.Offset != -1)
            {
                map.br.BaseStream.Position = Part.Offset;
                Part.Raw = map.br.ReadBytes(Part.Size);
            }
            else if (Part.RawLocation != RawLocation.Internal && Part.Offset != -1)
            {
                BinaryReader br = map.OpenMap(Part.RawLocation);
                br.BaseStream.Position = Part.Offset;
                Part.Raw = br.ReadBytes(Part.Size);
                br.Close();
            }
            Particles.Add(Part);
            xr.WriteEndElement();
            xr.Close();

            //Write Raw
            if (Part.Raw != null)
            {
                File.WriteAllBytes(Location + ".Raw", Part.Raw);
            }
        }
        #endregion
    }

    public static class WeatherRaw
    {
        public static List<RawChunk> Weather = new List<RawChunk>();

        #region Methods
        public static void Read(Map map)
        {
            Weather.Clear();
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset;
            Reflexive r = new Reflexive();
            r.Read(map, true);
            for (int i = 0; i < r.ChunkCount; i++)
            {
                RawChunk Weat = new RawChunk();
                map.br.BaseStream.Position = r.Translation + (i * 140) + 64;
                Weat.Offset = map.br.ReadInt32();
                Weat.Size = map.br.ReadInt32();
                Weat.RawLocation = RawData.RawLocator(ref Weat.Offset);
                if (Weat.RawLocation == RawLocation.Internal && Weat.Offset != -1)
                {
                    map.br.BaseStream.Position = Weat.Offset;
                    Weat.Raw = map.br.ReadBytes(Weat.Size);
                }
                else if (Weat.RawLocation != RawLocation.Internal && Weat.Offset != -1)
                {
                    BinaryReader br = map.OpenMap(Weat.RawLocation);
                    br.BaseStream.Position = Weat.Offset;
                    Weat.Raw = br.ReadBytes(Weat.Size);
                    br.Close();
                }
                Weather.Add(Weat);
            }
        }

        public static void Decompile(Map map, Tag tag, string Location)
        {
            Weather.Clear();
            //Set up Xml
            XmlWriter xr = XmlWriter.Create(Location + ".Rawxml");
            xr.WriteStartDocument();
            xr.WriteStartElement("Raw");

            map.br.BaseStream.Position = tag.MetaOffset;
            Reflexive r = new Reflexive();
            r.Read(map, true);
            xr.WriteElementString("RawChunks", r.ChunkCount.ToString());
            for (int i = 0; i < r.ChunkCount; i++)
            {
                RawChunk Weat = new RawChunk();
                xr.WriteStartElement("Chunk");
                map.br.BaseStream.Position = r.Translation + (i * 140) + 64;
                Weat.Offset = map.br.ReadInt32();
                xr.WriteElementString("Offset", Weat.Offset.ToString());
                Weat.Size = map.br.ReadInt32();
                xr.WriteElementString("Size", Weat.Size.ToString());
                Weat.RawLocation = RawData.RawLocator(ref Weat.Offset);
                xr.WriteElementString("Location", Weat.RawLocation.ToString());
                if (Weat.RawLocation == RawLocation.Internal && Weat.Offset != -1)
                {
                    map.br.BaseStream.Position = Weat.Offset;
                    Weat.Raw = map.br.ReadBytes(Weat.Size);
                }
                else if (Weat.RawLocation != RawLocation.Internal && Weat.Offset != -1)
                {
                    BinaryReader br = map.OpenMap(Weat.RawLocation);
                    br.BaseStream.Position = Weat.Offset;
                    Weat.Raw = br.ReadBytes(Weat.Size);
                    br.Close();
                }
                Weather.Add(Weat);
                xr.WriteEndElement();
            }
            xr.WriteEndElement();
            xr.Close();

            //Write Raw
            for (int i = 0; i < Weather.Count; i++)
            {
                //Write Raw
                BinaryWriter bw = new BinaryWriter(new FileStream(Location + ".Raw", FileMode.OpenOrCreate, FileAccess.Write, FileShare.Write));
                bw.BaseStream.Position = bw.BaseStream.Length;
                if (Weather[i].Raw != null)
                {
                    bw.Write(Weather[i].Raw);
                }
                bw.Close();
            }
        }
        #endregion
    }

    public static class AnimationRaw
    {
        public static List<RawChunk> Animation = new List<RawChunk>();

        #region Methods
        public static void Read(Map map)
        {
            Animation.Clear();
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 372;
            //Reflexive r = new Reflexive();
            //r.Read(map, true);
            //for (int i = 0; i < r.ChunkCount; i++)
            //{
                RawChunk Anim = new RawChunk();
                //map.br.BaseStream.Position = r.Translation + (i * 20) + 4;
                Anim.Size = map.br.ReadInt32();
                Anim.Offset = map.br.ReadInt32();
                Anim.RawLocation = RawData.RawLocator(ref Anim.Offset);
                if (Anim.RawLocation == RawLocation.Internal && Anim.Offset != -1)
                {
                    map.br.BaseStream.Position = Anim.Offset;
                    Anim.Raw = map.br.ReadBytes(Anim.Size);
                }
                else if (Anim.RawLocation != RawLocation.Internal && Anim.Offset != -1)
                {
                    BinaryReader br = map.OpenMap(Anim.RawLocation);
                    br.BaseStream.Position = Anim.Offset;
                    Anim.Raw = br.ReadBytes(Anim.Size);
                    br.Close();
                }
                Animation.Add(Anim);
            //}
        }

        public static void Decompile(Map map, Tag tag, string Location)
        {
            Animation.Clear();
            //Set up Xml
            XmlWriter xr = XmlWriter.Create(Location + ".Rawxml");
            xr.WriteStartDocument();
            xr.WriteStartElement("Raw");

            map.br.BaseStream.Position = tag.MetaOffset + 172;
            Reflexive r = new Reflexive();
            r.Read(map, true);
            xr.WriteElementString("RawChunks", r.ChunkCount.ToString());
            for (int i = 0; i < r.ChunkCount; i++)
            {
                RawChunk Anim = new RawChunk();
                xr.WriteStartElement("Chunk");
                map.br.BaseStream.Position = r.Translation + (i * 20) + 4;
                Anim.Offset = map.br.ReadInt32();
                xr.WriteElementString("Offset", Anim.Offset.ToString());
                Anim.Size = map.br.ReadInt32();
                xr.WriteElementString("Size", Anim.Size.ToString());
                Anim.RawLocation = RawData.RawLocator(ref Anim.Offset);
                xr.WriteElementString("Location", Anim.RawLocation.ToString());
                if (Anim.RawLocation == RawLocation.Internal && Anim.Offset != -1)
                {
                    map.br.BaseStream.Position = Anim.Offset;
                    Anim.Raw = map.br.ReadBytes(Anim.Size);
                }
                else if (Anim.RawLocation != RawLocation.Internal && Anim.Offset != -1)
                {
                    BinaryReader br = map.OpenMap(Anim.RawLocation);
                    br.BaseStream.Position = Anim.Offset;
                    Anim.Raw = br.ReadBytes(Anim.Size);
                    br.Close();
                }
                Animation.Add(Anim);
                xr.WriteEndElement();
            }
            xr.WriteEndElement();
            xr.Close();

            //Write Raw
            for (int i = 0; i < Animation.Count; i++)
            {
                //Write Raw
                BinaryWriter bw = new BinaryWriter(new FileStream(Location + ".Raw", FileMode.OpenOrCreate, FileAccess.Write, FileShare.Write));
                bw.BaseStream.Position = bw.BaseStream.Length;
                if (Animation[i].Raw != null)
                {
                    bw.Write(Animation[i].Raw);
                }
                bw.Close();
            }
        }
        #endregion
    }

    public static class CoconutRaw
    {
        public static List<RawChunk> Coconuts = new List<RawChunk>();

        #region Methods
        public static void Read(Map map)
        {
            Coconuts.Clear();
            map.br.BaseStream.Position = map.Tags[map.Index.TagCount - 1].MetaOffset + 80;
            Reflexive r = new Reflexive();
            r.Read(map, true);
            for (int i = 0; i < r.ChunkCount; i++)
            {
                RawChunk Coco = new RawChunk();
                map.br.BaseStream.Position = r.Translation + (i * 44) + 8;
                Coco.Offset = map.br.ReadInt32();
                Coco.Size = map.br.ReadInt32();
                Coco.RawLocation = RawData.RawLocator(ref Coco.Offset);
                if (Coco.RawLocation == RawLocation.Internal && Coco.Offset != -1)
                {
                    map.br.BaseStream.Position = Coco.Offset;
                    Coco.Raw = map.br.ReadBytes(Coco.Size);
                }
                else if (Coco.RawLocation != RawLocation.Internal && Coco.Offset != -1)
                {
                    BinaryReader br = map.OpenMap(Coco.RawLocation);
                    br.BaseStream.Position = Coco.Offset;
                    Coco.Raw = br.ReadBytes(Coco.Size);
                    br.Close();
                }
                Coconuts.Add(Coco);
            }
        }

        public static void Decompile(Map map, Tag tag, string Location)
        {
            Coconuts.Clear();
            //Set up Xml
            XmlWriter xr = XmlWriter.Create(Location + ".Rawxml");
            xr.WriteStartDocument();
            xr.WriteStartElement("Raw");

            map.br.BaseStream.Position = tag.MetaOffset + 80;
            Reflexive r = new Reflexive();
            r.Read(map, true);
            xr.WriteElementString("RawChunks", r.ChunkCount.ToString());
            for (int i = 0; i < r.ChunkCount; i++)
            {
                RawChunk Coco = new RawChunk();
                xr.WriteStartElement("Chunk");
                map.br.BaseStream.Position = r.Translation + (i * 44) + 8;
                Coco.Offset = map.br.ReadInt32();
                xr.WriteElementString("Offset", Coco.Offset.ToString());
                Coco.Size = map.br.ReadInt32();
                xr.WriteElementString("Size", Coco.Size.ToString());
                Coco.RawLocation = RawData.RawLocator(ref Coco.Offset);
                xr.WriteElementString("Location", Coco.RawLocation.ToString());
                if (Coco.RawLocation == RawLocation.Internal && Coco.Offset != -1)
                {
                    map.br.BaseStream.Position = Coco.Offset;
                    Coco.Raw = map.br.ReadBytes(Coco.Size);
                }
                else if (Coco.RawLocation != RawLocation.Internal && Coco.Offset != -1)
                {
                    BinaryReader br = map.OpenMap(Coco.RawLocation);
                    br.BaseStream.Position = Coco.Offset;
                    Coco.Raw = br.ReadBytes(Coco.Size);
                    br.Close();
                }
                Coconuts.Add(Coco);
                xr.WriteEndElement();
            }
            xr.WriteEndElement();
            xr.Close();

            //Write Raw
            for (int i = 0; i < Coconuts.Count; i++)
            {
                //Write Raw
                BinaryWriter bw = new BinaryWriter(new FileStream(Location + ".Raw", FileMode.OpenOrCreate, FileAccess.Write, FileShare.Write));
                bw.BaseStream.Position = bw.BaseStream.Length;
                if (Coconuts[i].Raw != null)
                {
                    bw.Write(Coconuts[i].Raw);
                }
                bw.Close();
            }
        }
        #endregion
    }

    public static class SoundRaw
    {
        public static List<RawChunk> Sounds = new List<RawChunk>();
        public static SoundPermutation Permutation = new SoundPermutation();

        #region Methods
        public static void Read(Map map)
        {
            Sounds.Clear();
            if (map.VerifySettings() == false)
            {
                MessageBox.Show("Error loading external map!", "Sound Raw Error");
                return;
            }
            Coconuts.Read(map);
            Permutation = new SoundPermutation();
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 8;
            int Index = map.br.ReadInt16();
            Permutation = Coconuts.Permutations[Index];
            for (int i = 0; i < Permutation.ChunkCount; i++)
            {
                int ChoiceIndex = Permutation.ChoiceIndex + i;
                Permutation.Choices.Add(Coconuts.Choices[ChoiceIndex]);
                for (int x = 0; x < Coconuts.Choices[ChoiceIndex].ChunkCount; x++)
                {
                    int SoundIndex = Coconuts.Choices[ChoiceIndex].SoundIndex + x;
                    Permutation.Choices[i].SoundChunks.Add(Coconuts.SoundChunks[SoundIndex]);
                    RawChunk Snd = new RawChunk();
                    Snd.Offset = Coconuts.SoundChunks[SoundIndex].Offset;
                    Snd.Size = Coconuts.SoundChunks[SoundIndex].Size; //& 0x3FFFFFFF
                    Snd.RawLocation = Coconuts.SoundChunks[SoundIndex].RawLocation;
                    Snd.Raw = Coconuts.SoundChunks[SoundIndex].Raw;
                    Sounds.Add(Snd);
                }
            }
        }

        public static void Decompile(Map map, Tag tag, string Location)
        {
            Sounds.Clear();
            if (map.VerifySettings() == false)
            {
                MessageBox.Show("Error loading external map!", "Sound Raw Error");
                return;
            }
            //Set up Xml
            XmlWriter xr = XmlWriter.Create(Location + ".Rawxml");
            xr.WriteStartDocument();
            xr.WriteStartElement("Raw");

            Permutation = new SoundPermutation();
            xr.WriteStartElement("Permutation");
            map.br.BaseStream.Position = tag.MetaOffset + 8;
            int Index = map.br.ReadInt16();
            Permutation = Coconuts.Permutations[Index];
            xr.WriteElementString("Index", Permutation.ChoiceIndex.ToString());
            xr.WriteElementString("Unknown1", Permutation.Unknown1.ToString());
            xr.WriteElementString("Unknown2", Permutation.Unknown2.ToString());
            xr.WriteElementString("Unknown3", Permutation.Unknown3.ToString());
            xr.WriteElementString("Unknown4", Permutation.Unknown4.ToString());
            xr.WriteElementString("ChunkNumber", Permutation.ChunkNumber.ToString());
            xr.WriteElementString("ChunkCount", Permutation.ChunkCount.ToString());
            for (int i = 0; i < Permutation.ChunkCount; i++)
            {
                xr.WriteStartElement("Choice");
                int ChoiceIndex = Permutation.ChoiceIndex + i;
                Permutation.Choices.Add(Coconuts.Choices[ChoiceIndex]);
                xr.WriteElementString("Index1", Coconuts.Choices[ChoiceIndex].NameIndex.ToString());
                xr.WriteElementString("SoundIndex", Coconuts.Choices[ChoiceIndex].SoundIndex.ToString());
                xr.WriteElementString("ChunkCount", Coconuts.Choices[ChoiceIndex].ChunkCount.ToString());
                xr.WriteElementString("Unknown1", Coconuts.Choices[ChoiceIndex].Unknown1.ToString());
                xr.WriteElementString("Unknown2", Coconuts.Choices[ChoiceIndex].Unknown2.ToString());
                for (int x = 0; x < Coconuts.Choices[ChoiceIndex].ChunkCount; x++)
                {
                    xr.WriteStartElement("Sound");
                    int SoundIndex = Coconuts.Choices[ChoiceIndex].SoundIndex + x;
                    Permutation.Choices[i].SoundChunks.Add(Coconuts.SoundChunks[SoundIndex]);
                    xr.WriteElementString("Offset", Coconuts.SoundChunks[SoundIndex].Offset.ToString());
                    xr.WriteElementString("Size", Coconuts.SoundChunks[SoundIndex].Size.ToString());
                    xr.WriteElementString("Location", Coconuts.SoundChunks[SoundIndex].RawLocation.ToString());
                    RawChunk Snd = new RawChunk();
                    Snd.Offset = Coconuts.SoundChunks[SoundIndex].Offset;
                    Snd.Size = Coconuts.SoundChunks[SoundIndex].Size; //& 0x3FFFFFFF
                    Snd.RawLocation = Coconuts.SoundChunks[SoundIndex].RawLocation;
                    Snd.Raw = Coconuts.SoundChunks[SoundIndex].Raw;
                    Sounds.Add(Snd);
                    xr.WriteEndElement();
                }
                xr.WriteEndElement();
            }
            xr.WriteEndElement();
            xr.Close();

            //Write Raw
            for (int i = 0; i < Sounds.Count; i++)
            {
                //Write Raw
                BinaryWriter bw = new BinaryWriter(new FileStream(Location + ".Raw", FileMode.OpenOrCreate, FileAccess.Write, FileShare.Write));
                bw.BaseStream.Position = bw.BaseStream.Length;
                if (Sounds[i].Raw != null)
                {
                    bw.Write(Sounds[i].Raw);
                }
                bw.Close();
            }
        }
        #endregion
    }
}
