﻿//*****************************************************************
// Created by Grimdoomer August 27, 2008
// Product of GrimProductions (c) 2008 - 2012
//
//*****************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace HaloMap
{
    public class RealTimeHalo
    {
        public static Process Halo;
        public static ProcessStream ps;

        public static bool Initialize(Map map)
        {
            Process[] Processes = Process.GetProcessesByName("Halo2");
            try { Halo = Processes[0]; }
            catch (Exception)
            {
                MessageBox.Show("Halo 2 is not running!");
                return false;
            }
            ps = new ProcessStream(map, Halo);
            ps.BaseAddress = map.SelectedTag.MetaOffset;
            return true;
        }
    }

	public class ProcessStream
    {
        #region Constructor
        /// <summary>
        /// Declares a new ProcessStream
        /// </summary>
        /// <param name="Process">Process to Open</param>
		public ProcessStream(Map map, Process Process)
		{
            this.m_ReadProcess = Process;
            this.map = map;
            m_hProcess = OpenProcess(PROCESS_VM_READ | PROCESS_VM_WRITE | PROCESS_VM_OPERATION, 1, (uint)m_ReadProcess.Id);
            byte[] buffer = new byte[4];
            ReadProcessMemory(m_hProcess, new IntPtr(MapVirtualOffset - 4), buffer, 4, out Out);
            VirtualIndexOffset = BitConverter.ToInt32(buffer, 0);
        }
        #endregion

        #region Properties
        /// <summary>	
		/// Active Process		
		/// </summary>
		public Process Process { get { return m_ReadProcess; } set { m_ReadProcess = value; } }
        /// <summary>
        /// Base Address ex: ProcessAddress + BaseAddress + Value Address
        /// </summary>
        public int BaseAddress { get { return baseaddress; } set { baseaddress = value; } }
        #endregion

        #region Members
        private Process m_ReadProcess = null;
        private IntPtr Out = IntPtr.Zero;
		private IntPtr m_hProcess = IntPtr.Zero;
        private int baseaddress = 0;
        public const int MapVirtualOffset = 18455360;
        private int VirtualIndexOffset;
        Map map;
        #endregion

        #region Methods
        public void Close()
		{
			int iRetValue;
			iRetValue = CloseHandle(m_hProcess);
			if (iRetValue == 0)
				throw new Exception("Real Time Halo Error: Close Handle Failed");
        }
        #endregion

        #region Reading
        public bool ReadBoolean(int Address)
        {
            byte[] buffer = new byte[4];
            ReadProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, 4, out Out);
            int Bool = BitConverter.ToInt32(buffer, 0);
            if (Bool > 0)
            {
                return true;
            }
            return false;
        }
        public byte ReadByte(int Address)
        {
            byte[] buffer = new byte[1];
            ReadProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, 1, out Out);
            return buffer[0];
        }
        public byte[] ReadBytes(int Address, int Count)
        {
            byte[] buffer = new byte[Count];
            ReadProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, (uint)Count, out Out);
            return buffer;
        }
        public char ReadChar(int Address)
        {
            byte[] buffer = new byte[0];
            ReadProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, 1, out Out);
            return BitConverter.ToChar(buffer, 0);
        }
        public char[] ReadChars(int Address, int Count)
        {
            byte[] buffer = new byte[Count];
            ReadProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, (uint)Count, out Out);
            char[] Chars = new char[Count];
            for (int i = 0; i < Count; i++)
            {
                Chars[i] = Convert.ToChar(buffer[i]);
            }
            return Chars;
        }
        public decimal ReadDecimal(int Address)
        {
            byte[] buffer = new byte[4];
            ReadProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, 4, out Out);
            return Convert.ToDecimal(buffer);
        }
        public double ReadDouble(int Address)
        {
            byte[] buffer = new byte[8];
            ReadProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, 8, out Out);
            return BitConverter.ToDouble(buffer, 0);
        }
        public short ReadInt16(int Address)
        {
            byte[] buffer = new byte[2];
            ReadProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, 2, out Out);
            return BitConverter.ToInt16(buffer, 0);
        }
        public int ReadInt32(int Address)
        {
            byte[] buffer = new byte[4];
            ReadProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, 4, out Out);
            return BitConverter.ToInt32(buffer, 0);
        }
        public long ReadInt64(int Address)
        {
            byte[] buffer = new byte[8];
            ReadProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, 8, out Out);
            return BitConverter.ToInt64(buffer, 0);
        }
        public sbyte ReadSByte(int Address)
        {
            byte[] buffer = new byte[1];
            ReadProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, 1, out Out);
            return Convert.ToSByte(buffer);
        }
        public float ReadSingle(int Address)
        {
            byte[] buffer = new byte[4];
            ReadProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, 4, out Out);
            return BitConverter.ToSingle(buffer, 0);
        }
        public string ReadString(int Address, int Length)
        {
            byte[] buffer = new byte[Length];
            ReadProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, (uint)Length, out Out);
            return BitConverter.ToString(buffer);
        }
        public ushort ReadUInt16(int Address)
        {
            byte[] buffer = new byte[2];
            ReadProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, 2, out Out);
            return BitConverter.ToUInt16(buffer, 0);
        }
        public uint ReadUInt32(int Address)
        {
            byte[] buffer = new byte[4];
            ReadProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, 4, out Out);
            return BitConverter.ToUInt32(buffer, 0);
        }
        public ulong ReadUInt64(int Address)
        {
            byte[] buffer = new byte[8];
            ReadProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, 8, out Out);
            return BitConverter.ToUInt64(buffer, 0);
        }
        #endregion

        #region Writing
        public void Write(int Address, byte Value)
        {
            WriteProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), BitConverter.GetBytes(Value), 1, out Out);
        }
        public void Write(int Address, byte[] Value)
        {
            WriteProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), Value, (uint)Value.Length, out Out);
        }
        public void Write(int Address, char Value)
        {
            WriteProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), BitConverter.GetBytes(Value), 1, out Out);
        }
        public void Write(int Address, char[] Value)
        {
            byte[] buffer = new byte[Value.Length];
            for (int i = 0; i < Value.Length; i++)
            {
                buffer[i] = Convert.ToByte(Value[i]);
            }
            WriteProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, (uint)Value.Length, out Out);
        }
        public void Write(int Address, decimal Value)
        {
            WriteProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), Misc.ConvertDecimalToByteArray(Value), 4, out Out);
        }
        public void Write(int Address, double Value)
        {
            WriteProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), BitConverter.GetBytes(Value), 8, out Out);
        }
        public void Write(int Address, short Value)
        {
            WriteProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), BitConverter.GetBytes(Value), 2, out Out);
        }
        public void Write(int Address, int Value)
        {
            WriteProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), BitConverter.GetBytes(Value), 4, out Out);
        }
        public void Write(int Address, long Value)
        {
            WriteProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), BitConverter.GetBytes(Value), 8, out Out);
        }
        public void Write(int Address, sbyte Value)
        {
            WriteProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), BitConverter.GetBytes(Value), 1, out Out);
        }
        public void Write(int Address, float Value)
        {
            WriteProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), BitConverter.GetBytes(Value), 4, out Out);
        }
        public void Write(int Address, string Value)
        {
            byte[] buffer = new byte[Value.Length];
            for (int i = 0; i < Value.Length; i++)
            {
                buffer[i] = Convert.ToByte(Value[i]);
            }
            WriteProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), buffer, (uint)Value.Length, out Out);
        }
        public void Write(int Address, ushort Value)
        {
            WriteProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), BitConverter.GetBytes(Value), 2, out Out);
        }
        public void Write(int Address, uint Value)
        {
            WriteProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), BitConverter.GetBytes(Value), 4, out Out);
        }
        public void Write(int Address, ulong Value)
        {
            WriteProcessMemory(m_hProcess, new IntPtr((Address + map.SelectedTag.RawMetaOffset) + VirtualIndexOffset), BitConverter.GetBytes(Value), 8, out Out);
        }
        #endregion

        #region Dll Imports
        // constants information can be found in <winnt.h>
        public const uint PROCESS_VM_READ = (0x0010);
        public const uint PROCESS_VM_WRITE = (0x0020);
        public const uint PROCESS_VM_OPERATION = (0x0008);

        // function declarations are found in the MSDN and in <winbase.h> 

        //		HANDLE OpenProcess(
        //			DWORD dwDesiredAccess,  // access flag
        //			BOOL bInheritHandle,    // handle inheritance option
        //			DWORD dwProcessId       // process identifier
        //			);
        [DllImport("kernel32.dll")]
        public static extern IntPtr OpenProcess(UInt32 dwDesiredAccess, Int32 bInheritHandle, UInt32 dwProcessId);

        //		BOOL CloseHandle(
        //			HANDLE hObject   // handle to object
        //			);
        [DllImport("kernel32.dll")]
        public static extern Int32 CloseHandle(IntPtr hObject);

        //		BOOL ReadProcessMemory(
        //			HANDLE hProcess,              // handle to the process
        //			LPCVOID lpBaseAddress,        // base of memory area
        //			LPVOID lpBuffer,              // data buffer
        //			SIZE_T nSize,                 // number of bytes to read
        //			SIZE_T * lpNumberOfBytesRead  // number of bytes read
        //			);
        [DllImport("kernel32.dll")]
        public static extern Int32 ReadProcessMemory(IntPtr hProcess, IntPtr lpBaseAddress, [In, Out] byte[] buffer, UInt32 size, out IntPtr lpNumberOfBytesRead);

        //		BOOL WriteProcessMemory(
        //			HANDLE hProcess,                // handle to process
        //			LPVOID lpBaseAddress,           // base of memory area
        //			LPCVOID lpBuffer,               // data buffer
        //			SIZE_T nSize,                   // count of bytes to write
        //			SIZE_T * lpNumberOfBytesWritten // count of bytes written
        //			);
        [DllImport("kernel32.dll")]
        public static extern Int32 WriteProcessMemory(IntPtr hProcess, IntPtr lpBaseAddress, [In, Out] byte[] buffer, UInt32 size, out IntPtr lpNumberOfBytesWritten);
        #endregion
    }
}
